#!/usr/bin/perl

use Date::Manip;

#
# Ad server cron script that runs once every hour to summarize ad stats so 
# far today just for this local web site. The local ad stats summary logs, ads_(date).log, 
# are then picked up at least once every hour by a cron process on the central log 
# archive server.
#
# Notes:
#	   - Loop foreach web site on this server? No, if script fails it would not write logs for the remaining web sites.
#          - Loop foreach logdate? Maybe someday, but would require consistent archiving of old access_log files.
#          - Why so many commandline arguments? So script can be cron'd on any server without hard-coding local configuration settings.
#          - Delete nfslogpath ads_siteid.serverid.log files? No, they will not accumulate more than number of serverid's.
#          - Is Date::Manip necessary? Only to calculate yesterday's logdate, otherwise I'd just use /bin/date
#
# Modified:  
#            Kurt Gray --                                  First rev. based on previous versions of ad_rotateset.pl
#            Kurt Gray -- Tue Oct 10 18:49:29 EDT 2000     Consolidating ads.log on NFS host
#            Kurt Gray -- Fri Oct 13 15:32:52 EDT 2000     Separate logging and rotation set scripts (ad_rotateset.pl and ad_makelog.pl)
#            Kurt Gray -- Thu Oct 26 11:36:43 EDT 2000     Adding commandline arguments to get rid of all hard-coded local configuartions
#            Kurt Gray -- Sun Nov  5 10:45:04 EST 2000     Added rotateaccesslog argument that date-copies access_log(s) then zero-out access_log(s)


# Read command line options

    foreach $arg (@ARGV) {

	# Option to specify a site

	if ( $arg =~ /site=(\w+)/ ) {
	    $siteid = $1;

	# Specify a base directory web sites are listed under

	} elsif ( $arg =~ /sites?dir=([\w\/]+)/ ) {
	    $sitesdir = $1;

	# Debug option shows what's happening without writing new logs
	# and a Ctrl+C keystroke will stop script and dumps stats gathered so far

	} elsif ( $arg =~ /debug/ ) {
	    $debugging = 1;
	    # $dontsavelogs = 1;
	    # $SIG{INT} = dumpStats;

	# Option to get stats for a specific date (date=yesterday is useful for regenerating yesterday's stats)

	} elsif ( $arg =~ /date=(\w+)/ ) {
	    $date_input = $1;

        # Number of lines to read in access_log before stopping (useful for debugging)

	} elsif ( $arg =~ /lines=(\d+)/ ) {
	    $lines_max = $1;

	# Option to start reading access_log at this line (debugging option, not implemented)

	} elsif ( $arg =~ /linestart=(\d+)/ ) {
	    $line_start = $1;

        # Option to run script wihtout saving logs (useful when debugging)

	} elsif ( $arg =~ /dontsavelogs/ ) {
	    $dontsavelogs = 1;

	# Option to specify base path of local access_log file

	} elsif ( $arg =~ /weblogpath=([\w\/\-\.]+)/ ) {
	    $weblogpath = $1;

	# Option to specify the full path of the access_log file that has ad impressions and clicks

	} elsif ( $arg =~ /accesslogfile=([\w\/\-\.]+)/ ) {
	    $accesslogfile = $1;

	# Option to rotate access_log and restart web server just at around midnight

	} elsif ( $arg =~ /rotateaccesslogs?/ ) {
	    $rotateaccesslog = 1;

	# Specify the path to write ad_(date).log files to

	} elsif ( $arg =~ /\badlogpath=([\w\/\-\.]+)/ ) {
	    $adlogpath = $1;

	# Option to consolidate ads_(date).log from several servers in a shared directory

	} elsif ( $arg =~ /nfsa?d?logpath=([\w\/\-\.]+)/ ) {
	    $nfsadlogpath = $1;

	# Specify a serverid so we can differentiate ad logs among NFS shared servers

	} elsif ( $arg =~ /serverid=([\w\-]+)/ ) {
	    $serverid = $1;

	# Base pattern to parse banner clicks in access_log

	} elsif ( $arg =~ /adclickpattern=([\w\.\/\-]+)/ ) {
	    $adclickpattern = $1;

	# Base pattern to parse banner impressions in access_log

	} elsif ( $arg =~ /adbannerpattern=([\w\/\-]+)/ ) {
	    $adbannerpattern = $1;

	} elsif ( $arg =~ /help/ ) {

	    print <<HELP_SCREEN;

$0 parses local web access_log file for local web server 
and creates and ads_(date).log file.

Common Usage (Single Web Server, Local Log Path): 

    $0  site=xx000en  
        weblogpath=/usr/local/apache/sites/mysite.com/logs 


Usage On One of Many Load-Balancing Web Servers (NFS Mode):

    $0  site=xx000en  
        weblogpath=/usr/local/apache_logs/mysite.com/logs
        nfsadlogpath=/usr/local/apache/logs  
        serverid=cpu54

When Local access_log and ads_(date).log Files Are In Different Paths:

    $0  site=xx000en  
        accesslogfile=/usr/local/apache/logs/www.mysite.com.accesslog
        adlogpath=/usr/local/apache/sites/mysite.com/log


Rotate Local access_log File(s) Then Quit (must run as root):

    $0  rotateaccesslog
        weblogpath=/usr/local/apache/sites/mysite.com/logs


Regenerating Ad Stats for Yesterday:

    $0  yesterday  
        site=xx000en  
        weblogpath=/usr/local/apache/sites/mysite.com/logs


Regenerating Ad Stats for an Arbitrary Date:

    $0  date=001106  (Note: date format can be yyMMDD or YYYYMMDD)
        site=xx000en  
        adlogpath=/usr/local/apache/sites/mysite.com/logs
        accesslogfile=/usr/local/apache/sites/logs/access_log.Mon


Specifying Base URI Patterns to Parse access_log for Ad Impressions and Clicks:

    $0  site=xx000en  
        weblogpath=/usr/local/apache/sites/mysite.com/logs
        adbannerpattern=/adimages/  
        adclickpattern=ad_click.pl


Debugging Options:

    $0 ... debug ...         (Verbose Ouput Mode)

    $0 ... dontsavelogs ...  (No Files are Created, Renamed, Overwritten, or Saved)


Author:

    Kurt Gray at OSDN.com

Version:

    Nov 5, 2000

HELP_SCREEN

            exit;

	# Warn about unparsed inputs
	} else {
	    print "ERROR: $0 does not understand argument: $arg \n";
	    exit;
	} # if arg

    }  # foreach ARGV



# Determine the datestamp part of the ad log filename (logdate)

    if ($date_input eq 'yesterday') {

	( $logdate, $dayofweek ) = split ',', &Date::Manip::UnixDate("yesterday", '%y%m%d,%a');
	$accesslogfiledate = $dayofweek; # dayofweek is %a format = 'Mon', 'Tue', 'Wed', etc


    } elsif ( $date_input =~ /(\d{6})$/ ) {

        $logdate = $date_input;

	if ( $accesslogfile !~ /\w+/ ) {
	    print "ERROR: When specifying a logdate must also specify access_log file name:\n";
	    print "Example: $0 date=20001024 accesslogfile=/usr/logs/old/access_log.Wed\n";
	    exit;
	}

    } else {

	$hour = `/bin/date +'%H'`;	    # If it's after midnight, then redo yesterday's stats
	if ( $hour =~ '00' ) {

	    ( $logdate, $dayofweek ) = split ',', &Date::Manip::UnixDate("yesterday", '%y%m%d,%a');
	    $accesslogfiledate = $dayofweek;

	} else {
	    ( $logdate, $dayofweek ) = split ',', &Date::Manip::UnixDate("today", '%y%m%d,%a');
	}

    } # if date_input


# Check mandatory command line arguments:

    if ( $rotateaccesslog == 1 ) {

	if ( $weblogpath =~ /\w+/ ) {
	    print "DEBUG: rotateaccesslog == 1 : Going to backup access_log(s) then zero-out access_log(s) \n" if $debugging == 1;
	} else {
	    print "ERROR: Must specify the weblogpath when using rotateaccesslog feature:\n";
	    print "Example: $0 weblogpath=/usr/local/apache/sites/mysite.com/logs rotateaccesslog ...\n";
	    exit;
	}

    } else {

	# siteid should be something like 'sh000en' which becomes the fourth column in the ad log

	if ( $siteid =~ /^\w{7}$/ ) {
	    print "DEBUG: siteid preset to '$siteid' \n" if $debugging == 1;
	} else {
	    print "ERROR: Must specify a local siteid in command line:\n";
	    print "Example: $0 site=xx000en ... \n";
	    exit;
	} # if siteid

} # if rotateaccesslog

    # If an nfsadlogpath is specified then this script will also consolidate ad logs from various serverid's and siteid's

    if ( $nfsadlogpath =~ /\w+/ ) {
	print "DEBUG: NFSMODE: nfsadlogpath preset to '$nfsadlogpath' \n" if $debugging == 1;
	if ( $serverid =~ /\w+/ ) {
	    print "DEBUG: NFS Mode: local serverid preset to '$serverid' \n" if $debugging == 1;
	    $nfsadlogfile = "$nfsadlogpath/ads\_$siteid\.$serverid\.log";
	} else {
	    print "ERROR: When specifying nfsadlogpath must also specify the local serverid:\n";
	    print "Example: $0 nfsadlogpath\=$nfsadlogpath serverid=(serverid) ... \n\n";
	    exit;
	}
    } else {
	print "DEBUG: Doing single server mode since nfsadlogpath was not specified. \n" if $debugging == 1;
    }


# Script default settings:

    $adbannerpattern   ||=   '/banner/';
    $adclickpattern    ||=   'adlog.pl';
    $sitesdir          ||=   '/usr/local/apache/sites';
    $weblogpath        ||=   "$sitesdir/$siteid/logs";
    $accesslogfile     ||=   "$weblogpath/access_log";
    $accesslogfile     .=    "\.$accesslogfiledate" if $accesslogfiledate; 
    $adlogpath         ||=   $weblogpath;
    $adlogfile         ||=   "$adlogpath/ads\_$logdate.log\.$siteid";
    $adclicklimit      ||=   100;     # Maximum number of ad clicks counted from a single IP address (a little crazy web-bot protection)


    if ( $debugging == 1 ) {
	$time_start = time;
	print "DEBUG: ================= VERIFY THE SETTINGS MARKED WITH ASTERICKS ===============\n"; 
	print "DEBUG: Base path of this web site             sitesdir          '$sitesdir'\n";
	print "DEBUG: Base path of local access_log file     weblogpath        '$weblogpath'\n";
	print "DEBUG: Datestamp of local access_log file     accesslogfiledate '$accesslogfiledate'\n";
	print "DEBUG: Full path to local access_log file     accesslogfile  *  '$accesslogfile'\n";
	print "DEBUG: Path to write ads_date.log file to     adlogpath         '$adlogpath'\n";
	print "DEBUG: Full path to write local adlogfile     adlogfile      *  '$adlogfile'\n";
	print "DEBUG: Base pattern for banner clicks         adclickpattern    '$adclickpattern'\n";
	print "DEBUG: Max number of ad clicks from same IP   adclicklimit      '$adclicklimit'\n";
	print "DEBUG: Base pattern for banner impressions    adbannerpattern   '$adbannerpattern'\n\n";
    }



# If rotateaccesslog == 1 then rotate each access_log in weblogpath and quit
# Note: We do this at this point in the script because we needed to know what the accesslogfiledate is.

    if ( $rotateaccesslog == 1 ) {
	&rotateAccessLog;
	exit;
    }

# Get today's ad stats for this web site (kept as a subroutine in case we want to repeat it with multiple siteids)

    %AdImpressions = ();
    %AdClicks = ();
    %AdClicksPerVisitor = ();
    $adlogdata = '';

    &readAccessLog;

# Save today's ad stats for this local web site (siteid) to a local ad stats log file

    $adlogdata_size = length $adlogdata;
    if ( $adlogdata_size < 4 ) {
	print "NOTICE: Not saving $adlogfile : ad stats data empty. Perhaps there were no ad impressions in $accesslogfile \n";
	exit;
    }


    if ( $dontsavelogs == 1 ) {
	print "NOTICE: Not actually saving $adlogfile ($adlogdata_size bytes) : dontsavelogs == 1 \n";
    } else {
	if ( open ( OUTFILE, "> $adlogfile" ) ) {
	    print OUTFILE $adlogdata;
	    close OUTFILE;
	    print "DEBUG: Saved adlogdata to $adlogfile : $adlogdata_size bytes written. \n" if $debugging == 1;
	} else {
	    print "ERROR: Could not write to local file $adlogfile: $! \n";
	}
    }


# Done for now is no nfsadlogpath was specified

    if ( $nfsadlogfile !~ /\w+/ ) {
	print "DEBUG: $0 Single Server Mode: Done.\n" if $debugging == 1;
	exit;
    }


# Save local ad stats for this siteid.serverid to NFS host in siteid.serverid ad stats temp file

    if ( $dontsavelogs == 1 ) {
	print "NOTICE: Not actually saving $nfsadlogfile ($adlogdata_size bytes) : dontsavelogs == 1 \n";
    } else {
	if ( open ( OUTFILE, "> $nfsadlogfile" ) ) {
	    print OUTFILE "logdatestamp\=$logdate\n";    # The first line of nfsadlogfile temp file is the datestamp
	    print OUTFILE $adlogdata;
	    close OUTFILE;
	    print "DEBUG: Saved adlogdata to $nfsadlogfile : $adlogdata_size bytes written. \n" if $debugging == 1;
	} else {
	    print "ERROR: Could not write to NFS file $nfsadlogfile: $! \n";
	    exit; # No sense going further if writing to NFS mount failed
	}
    }

# Clear out any previously used ad data arrays and variables

   $adlogdata      = '';
   %AdImpressions  = ();
   %AdClicks       = ();
   %AllAds         = ();
   %AllSites       = ();

   print "DEBUG: Looking for siteid.serverid ad stats temp files in $nfsadlogpath ...\n" if $debugging == 1;

   if (opendir (DIR, "$nfsadlogpath")) {
       @nfsadlogfiles = grep /^ads\_\w{7}\.[\w\-]+\.log/, (readdir DIR);
       close DIR;	    
   } else {
       print "ERROR: $0 can not opendir $nfsadlogpath : $!";
       exit;
   } 

# Read all of the siteid.serverid temp ad logs and consolidate all ad stats for each siteid for this logdate

   foreach $nfsadlogfile (sort @nfsadlogfiles) {
       print "\nDEBUG: Consolidating NFS logs: Looking at $nfsadlogfile...\n" if $debugging == 1;
       if ( open INFILE, "< $nfsadlogpath/$nfsadlogfile" ) {
	   print "DEBUG: Reading $nfsadlogfile ... \n" if $debugging == 1;
	   $numberofnfsfiles++;
	       
	   # First line of nfs temp log should be a datestamp: skip this file if its datestamp is not logdate

	   $tmp_datestamp = <INFILE>;

	   if ( $tmp_datestamp !~ /logdatestamp\=(\d{6})/ ) {
	       print "ERROR: $nfsadlogfile does not have a logdatestemp in top line. \n";
	       close INFILE;
	       next;
	   } else {
	       $logdatestamp = $1;
	       if ( $logdatestamp ne $logdate ) {
		   print "DEBUG: Skipping $nfsadlogfile : logdatestamp, $logdatestamp != logdate, $logdate\n" if $debugging == 1;
		   close INFILE;
		   next;
	       }
	   }
	       
	   while (<INFILE>) {
	       chomp;
	       ($AdID, $impressions, $clicks, $tmp_siteid) = split(/\t/, $_);
	       $AllAds{$AdID}++;
	       $AllSites{$tmp_siteid}++;
	       $AdImpressions{$AdID . $tmp_siteid} += $impressions;
	       $AdClicks{$AdID . $tmp_siteid} += $clicks;
	   }

	   close INFILE;

       } else {
	   print "ERROR : Could not read $nfsadlogfile : $! \n";
       }
       
   } # foreach nfsadlogfile


# Now dump the reconsolidated ad stats for this logdate, all sites, all web servers into one ad stats file

   foreach $tmp_siteid (sort keys %AllSites) {
       foreach $AdID (sort keys %AllAds) {
	   $number_of_clicks      =  $AdClicks{       $AdID . $tmp_siteid } * 1;
	   $number_of_impressions =  $AdImpressions{  $AdID . $tmp_siteid } * 1;
	   next if  $number_of_impressions < 1;
	   $adlogdata .= "$AdID\t$number_of_impressions\t$number_of_clicks\t$tmp_siteid\n";
       }
   }

   $adlogdata_size = length $adlogdata;
   $nfsadlogfile = "$nfsadlogpath/ads\_$logdate.log";
       
   if ( $adlogdata_size < 4 ) {
       print "NOTICE: Not saving $nfsadlogfile since we didn't find any log temp files in $nfsadlogapth. \n";
       exit;
   }

   if ( $dontsavelogs == 1 ) {
       print "NOTICE: Stopping before saving $nfsadlogfile : dontsavelogs == 1 \n";
       exit;
   }


   if ( $dontsavelogs == 1 ) {
       print "NOTICE: Not actually saving $nfsadlogfile ($adlogdata_size bytes) : dontsavelogs == 1 \n";
   } else { 
       if ( open ( OUTFILE, "> $nfsadlogfile" ) ) {
	   print OUTFILE $adlogdata;
	   close OUTFILE;
	   print "DEBUG: Saved adlogdata to $nfsadlogfile : $adlogdata_size bytes written. \n" if $debugging == 1;
       } else {
	   print "ERROR: Could not write to NFS file $nfsadlogfile: $! \n";
       }
   }
	    
   print "DEBUG: $0 Finished NFS ad log consolidation. Done.\n" if $debugging == 1;
   exit;






# Count number of ad impressions and ad clicks per ad from local the access_log

sub readAccessLog {

    if ( open(INFILE, "< $accesslogfile") ) {

	print "DEBUG: Parsing lines in $accesslogfile ...\n" if $debugging == 1;

	while (<INFILE>) {

	    next if ( $line_number++ < $line_start ) && ( $line_start > 1);
	    last if ( $line_number > $lines_max ) && ( $lines_max > 1 );

	    if ( $_ !~ /^(\S+) \- \- \[.+\] \"([A-Z])+ (\S+) \S+\" [23]/ ) {
		# print "DEBUG: LINE REJECTED: $_ " if $debugging == 1;
		next;
	    }

	    $remote_host = $1;
	    next if $2 eq 'HEAD';
	    $request = $3;
	    
	    if ($request =~ /$adbannerpattern(\w{10})\.gif/) {
		$total_impressions++;
		$AdImpressions{$1}++;
	    } elsif ($request =~ /$adclickpattern\?([\w\.\-\/]*),(\w{10})/) {
		if ( $adclicklimit > 1 ) {
		    if ( $AdClicksPerVisitor{$remote_host}++ > $adclicklimit ) {
                        print "DEBUG: CLICK REJECTED: $2: Ad clicks for $remote_host exceeds $adclicklimit \n" if $debugging == 1;
                        next;
		    }
		}
		$total_clicks++;
		$AdClicks{$2}++;
	    } else {
		next;
	    }

	}  # while access_log
	close INFILE;

	foreach $AdID (sort keys %AdImpressions) {
	    $number_of_ads++;
	    $number_of_clicks = $AdClicks{$AdID} * 1;
	    $adlogdata .= "$AdID\t" . $AdImpressions{$AdID}  . "\t" . $number_of_clicks . "\t$siteid\n";
	}

	if ( $debugging == 1 ) {
	    $number_of_ads *= 1;
	    $time_elapsed = time - $time_start;
	    print "DEBUG: Done reading $accesslogfile : $line_number lines read in $time_elapsed seconds:\n";
	    print "       $number_of_ads ads found, $total_impressions impressions, $total_clicks clicks.\n";
	}

    } else {

	print "ERROR: Can't read $accesslogfile: $! \nHINT: Verify your weblogpath= or accesslogfile= argument.\n";
	
    } # if acccesslogfile

 }  # sub readAccessLog





# Find all access_log(s) in weblogpath, make day-of-week backup copy, then zero-out access_log(s)

sub rotateAccessLog {

    # Only root user can do this
    if ( $> != 0 ) {
	print "ERROR: Only the root user can effectively use the rotateaccesslog option. ( Effective uid = $> ) \n";
	return;
    }

    print "DEBUG: Looking in $weblogpath for access_logs to rotate...\n" if $debugging == 1;

    $accesslogfiledate = $dayofweek;
    $accesslogfilename = '';
    @accesslogfiles    = ();
    @logfiles          = ();


    if (opendir (DIR, "$weblogpath")) {
	@logfiles = grep /log$/, (readdir DIR);
	close DIR;

	     @accesslogfiles = grep /access/,  @logfiles;   # Tried doing this all in one grep but it matched files I didn't want
	push @accesslogfiles, (grep /error/,   @logfiles);
	push @accesslogfiles, (grep /referer/, @logfiles);
	push @accesslogfiles, (grep /agent/,   @logfiles);


	foreach $accesslogfilename ( sort @accesslogfiles ) {

	    $accesslogfile = "$weblogpath/$accesslogfilename";
		    
	    print "\nDEBUG: $accesslogfilename is ", -s "$accesslogfile", " bytes before rotating.\n" if $debugging == 1; 

	    if ( $dontsavelogs == 1 ) {
		print "NOTICE: Not actually rotating $accesslogfilename ==> $accesslogfilename\.$accesslogfiledate : dontsavelogs == 1 \n";
		$num_files++;
		next;
	    }

	    # By copying the current access_log to new filename then zero out access_log we don't have to restart Apache

	    if ( open INFILE, "< $accesslogfile" ) {
		if ( open OUTFILE, "> $accesslogfile\.$accesslogfiledate" ) {
		    print "DEBUG: Copying $accesslogfile ==> $accesslogfile\.$accesslogfiledate\n" if $debugging == 1;
		    while (<INFILE>) {
			print OUTFILE;
		    }
		    close OUTFILE;

		    qx|/bin/cat /dev/null >  $accesslogfile|;

		    print "DEBUG: $accesslogfilename is now ", -s "$accesslogfile", " bytes after rotating.\n" if $debugging == 1; 
		    $num_files++;

		} else {
		    print "ERROR: Could not write to $accesslogfile\.$accesslogfiledate : $! \n";
		}
		close INFILE;
	    } else {
		print "ERROR: Could not read to $accesslogfile : $! \n";
	    }

	} # foreach rotate access_log
 
	if ( $num_files == 1 ) {
	    print "DEBUG: $accesslogfile has been rotated.\n" if $debugging == 1;
	} elsif ( $num_files > 1 ) {
	    print "DEBUG: $num_files access_log files rotated.\n" if $debugging == 1;
	} else {
	    print "ERROR: No access_log files found in $weblogpath.\n";
	}

    } else {
	print "ERROR: RotateAccessLogs: Could not read directory $weblogpath : $! \n";
    }

    return;

} # sub rotateAccessLog





__END__









