#!/usr/bin/perl

use Date::Manip;

#
# This shell script runs once every hour to log
# all ad stats so far today then adjusts new ad
# rotation lists according to ad_x_alloc.dat.
#
# Notes:
#	- Data beyond 50 clicks from a single IP is ignored (assumed to be a robot)
#	- Ads are stopped for rest of day if StopOnTarget 
#	- Ads are stopped on given StopDate
#	- Sets position 1 and 4 ad rotation
#	- Added command line option to generate yesterday's log file
#	- Creates a lean version of ad_info.dat (ad_info_lite.dat)
#       - Commented out the ErrorDocument line to allow http.conf directive to work. (Marc)
#
# Modified:  Kurt Gray -- Thu Mar 23 09:37:29 EST 2000	  Don't process blank adids
#            Kurt Gray -- Tue Oct 10 18:49:29 EDT 2000    Consolidating ads.log on NFS host
#            Kurt Gray -- Fri Oct 13 15:32:52 EDT 2000    Separate logging and set rotation scripts (ad_rotateset.pl and ad_makelog.pl)
#            Kurt Gray -- Fri Oct 20 15:44:41 EDT 2000    Modified to be centralized on report.andover.net
#            Kurt Gray -- Mon Nov  6 11:21:38 EST 2000    Trying to centralize adsystem under /usr/local/adsystem on report and all sites

push @INC, '/usr/local/adsystem/lib/';
require 'ad_config_lib.pl';	        # Site-specific settings for adsystem control scripts


# Read command line options

    foreach $arg (@ARGV) {

	# Option to specify a site

	if ( $arg =~ /site=(\w+)/ ) {
	    $siteid = $1;

	# Debug option shows what's happening without writing new logs
	# and a Ctrl+C keystroke will stop script and dumps stats gathered so far

	} elsif ( $arg =~ /debug/ ) {
	    $debugging = 1;
	    # $dontsavelogs = 1;
	    # $SIG{INT} = dumpStats;

	# Option to get stats for a specific date (date=yesterday is useful for regenerating yesterday's stats)

	} elsif ( $arg =~ /date=(\w+)/ ) {
	    $date_input = $1;

        # Number of lines to read in access_log before stopping (useful for debugging)

	} elsif ( $arg =~ /lines=(\d+)/ ) {
	    $lines_max = $1;

	# Specify a base directory web sites are listed under

	} elsif ( $arg =~ /sitesdir=([\w\/]+)/ ) {
	    $sitesdir = $1;

	# Option to start reading access_log at this line (debugging option, not implemented)

	} elsif ( $arg =~ /linestart=(\d+)/ ) {
	    $line_start = $1;

        # Option to run script wihtout saving logs (useful when debugging)

	} elsif ( $arg =~ /dontsavelogs/ ) {
	    $dontsavelogs = 1;

	# Option to write number_of_hits per IP column in visitors_(date).log

	} elsif ( $arg =~ /positions=([\d,]+)/ ) {
	    @positions_list_input = split(/,/, $1);

	# Option to ignore pagecount gif and do regular hit parsing

	} elsif ( $arg =~ /nopagecountgif/ ) {
	    $nopagecountgif = 1;

	# Warn about unparsed inputs
	} else {
	    print "ERROR: $0 does not understand argument: $arg \n";
	    exit;
	} # if arg


    }


    $logdate = $date_input;
    $logdate ||= &Date::Manip::UnixDate("today", '%y%m%d');
    $todayfullsysdate = &Date::Manip::UnixDate("today", '%Y%m%d');


    $hour = `/bin/date +'%H'`;	# If it's after midnight, then redo yesterday's stats

    if (($date_input eq 'yesterday') || ($hour =~ '00')) {
	$logdate = &Date::Manip::UnixDate("yesterday", '%y%m%d');
    }


    $sitesdir   ||= "/usr/local/apache/sites/report/logs";
   
    $adlogfile  ||= "$adlogpath/ads\_$logdate.log";



# Consolidate ad stats into an array, key siteid.AdID => impressions (used for checking StopOnTarget limited ads)

    &readAdStats;

# Read the master ad_info.dat file and key each AdID => height, width, target, etc. (used when creating ad_info2.dat)

    &readAdInfo;


# Make a global list of all siteid's

    foreach $logname ( sort keys %logname2siteid ) {
	push @allsiteids, $logname2siteid{$logname};
    }


# Decide whether to do this for an individual web site or all sites listed in ad_config_lib.pl

    if ( $siteid ) {

	&readSiteConfig;
	print "\nDEBUG: SINGLE SITE MODE: Looking at $siteid ($sitename) ...\n" if $debugging == 1;
	&setAdRotations;

    } else {

	foreach $siteid ( sort @allsiteids ) {
	    next if $siteid !~ /\w\w000en/;
	    &readSiteConfig;
	    print "\nDEBUG: Looking at $siteid ($sitename) ...\n" if $debugging == 1;
	    $numberofsites++;
	    &setAdRotations;
	}

    } # if siteid


    if ( $system_logfile ) {
	my $date = `/bin/date`; chomp $date;
	my $dow = $1 if $date =~ /^(\w{3}) /;
	$system_logfile = "$system_logfile.$dow";
	if ( ( -M "$system_logfile" ) > 3 ) {
	    `/bin/cp /dev/null $system_logfile`;  # Zero out log on date roll over
	    }
	if ( open LOGFILE, ">> $system_logfile" ) {
	    $numofsites_rotated = @rotated_sites;
	    print LOGFILE qq|$date\t[adrotateset]\t$numberofsites sites listed\t$numofsites_rotated sites rotated\t|;
	    print LOGFILE ( join ', ' , @rotated_sites ) if $debugging == 1;
	    print LOGFILE "\n";
	    close LOGFILE;
	}
    } # logfile



# End of main




# Any special ad system behaviour with this web site comes from ad_config_lib.pl

sub readSiteConfig {

    $site_initials       = substr( $siteid, 0, 2 );
    $sitename            = $sitesconfig{ $siteid . 'sitename'            };
    $ad_status           = $sitesconfig{ $siteid . 'ad_status'           };
    $ad_allocation_group = $sitesconfig{ $siteid . 'ad_allocation_group' };
    $adinfolite          = $sitesconfig{ $siteid . 'ad_info_lite'        } || 'ad_info2.dat';

    # Make a list of allocation subsites of this siteid

    @ad_allocation_subsites = ( $siteid );
    foreach $tmp_siteid ( sort @allsiteids ) {
	next if $tmp_siteid eq $siteid;
	if (  $sitesconfig{ $tmp_siteid . 'ad_allocation_group' } eq $site_initials ) {
	    push @ad_allocation_subsites, $tmp_siteid;
	}
    }


    # Allow shell user to overide positions list in command line

    if ( @positions_list_input ) {
	@positions_list = @positions_list_input;
    } else {
	@positions_list = split(/,/, $sitesconfig{$siteid.'ad_positions'});
    }

    print "WARNING: siteid '$siteid' does not have a sitename in \%sitesconfig. \n" if !$sitename;

    if ( $debugging == 1 ) {
	print "\nDEBUG: \%sitesconfig $siteid ad_status            = '$ad_status' \n";
	print "DEBUG: \%sitesconfig $siteid ad_allocation_group       = '$ad_allocation_group' \n";
	print "DEBUG: \%sitesconfig $siteid \@positions_list          = '", @positions_list, "' \n";
	print "DEBUG: \%sitesconfig $siteid \@ad_allocation_subsites  = '", @ad_allocation_subsites, "' \n";
    }

} # end readSiteConfig



sub setAdRotations {

    %CurrentAdsList            = ();
    $reserved_filedata         = '';
    $reserved_domains_filedata = '';
    $ad_info_data              = '';
    

    if ( $ad_status eq 'ACTIVE' ) {

    } else {
	print "DEBUG: Skipping MakeRotateFile for $siteid: ad_status = '$ad_status' \n" if $debugging == 1;
	return;
    }


# Make new rotation file for each ad position


    if ( $ad_allocation_group ) {
	print "DEBUG: Skipping makeRotateFile for $siteid: ad_allocation_group = $ad_allocation_group \n" if $debugging == 1;
	return;
    } else {
	if ( @positions_list ) {
	    foreach $ad_position (@positions_list) {
		&makeRotateFile;
	    }
	    push @rotated_sites, "$siteid:" . ( join "", @positions_list );
	} else {
	    print "ERROR: ad positions_list was empty for $siteid \n";
	}
    }



# Generate reserved pages file ad_reservations.dat by reading ad_reservations_alloc.dat

    if ( open (INFILE, "< $datfilepath/ad_reservations_alloc.dat\.$siteid" ) ) {
	while (<INFILE>) {
	    chomp;
	    
	    ($pagename, $position, $AdID, $StartDate, $StopDate, $allocated_impressions) = split(/\t/, $_);
		    
	    next if $AdID !~ /\w+/;
	    
	    # Stop reserved ad if daily impressions already delivered:

	    if ( $allocated_impressions > 10 ) {
		$impressions_delivered = $AdImpressions{ $siteid . $AdID };
		if ( $debugging == 1 ) {
		    print "DEBUG: $siteid Reserved Page $pagename : $AdID StopOnTarget, ";
		    print "reserved allocation = $allocated_impressions, impressions_delivered = $impressions_delivered : ";
		}
		if ( $impressions_delivered >= $allocated_impressions ) {
		    print "ad stopped.\n" if $debugging == 1;
		    next;
		} else {
		    print "ad allocated again.\n" if $debugging == 1;
		}
	    }
		    
	    # Consider StartDate and StopDate of this reserved ad:

	    if ( $StartDate ) {
		$StartDate = &ISODate2SysDate($StartDate);
		print "DEBUG: Comparing ReservedPage $AdID todayfullsysdate = $todayfullsysdate, StartDate ==> $StartDate : " if $debugging == 1;
		if ( $todayfullsysdate < $StartDate ) {
		    print "ad not started yet. \n" if $debugging == 1;
		    next;
		} else {
		    print "ad has started running. \n" if $debugging == 1;
		}
	    }
	    if ( $StopDate ) {
		$StopDate = &ISODate2SysDate($StopDate);
		print "DEBUG: Comparing ReservedPage $AdID todayfullsysdate = $todayfullsysdate, StopDate ==> $StopDate, " if $debugging == 1;
		if ( $todayfullsysdate >= $StopDate ) {
		    print "ad stopped on date. \n" if $debugging == 1;
		    next;
		} else {
		    print "ad still running. \n" if $debugging == 1;
		}
	    }
		    
	    $CurrentAdsList{$AdID} = 1;
		    
	    $reserved_filedata .= "$pagename\t$position\t$AdID\n";
		    
	}
	close INFILE;
	
    } # if ad_reservations_alloc.dat


# Generate reserved domains file ad_reserved_domains.dat by reading ad_reserved_domains_alloc.dat

    if ( open (INFILE, "< $datfilepath/ad_reserved_domains_alloc.dat\.$siteid" ) ) {
	while (<INFILE>) {
	    chomp;
	    
	    ($domainname, $position, $AdID, $StartDate, $StopDate, $allocated_impressions) = split(/\t/, $_);
		    
	    next if $AdID !~ /\w+/;

	    # Stop reserved ad if daily impressions already delivered:

	    if ( $allocated_impressions > 10 ) {
		$impressions_delivered = $AdImpressions{ $siteid . $AdID };
		if ( $debugging == 1 ) {
		    print "DEBUG: $siteid Reserved Domain $domainname : $AdID StopOnTarget, ";
		    print "reserved allocation = $allocated_impressions, impressions_delivered = $impressions_delivered : ";
		}
		if ( $impressions_delivered >= $allocated_impressions ) {
		    print "ad stopped.\n" if $debugging == 1;
		    next;
		} else {
		    print "ad allocated again.\n" if $debugging == 1;
		}
	    }

	    # Consider StartDate and StopDate of this reserved ad

	    if ( $StartDate ) {
		$StartDate = &ISODate2SysDate($StartDate);
		print "DEBUG: Comparing ReservedDomain $AdID todayfullsysdate = $todayfullsysdate, StartDate ==> $StartDate : " if $debugging == 1;
		if ( $todayfullsysdate < $StartDate ) {
		    print "ad not started yet. \n" if $debugging == 1;
		    next;
		} else {
		    print "ad has started running. \n" if $debugging == 1;
		}
	    }
	    if ( $StopDate ) {
		$StopDate = &ISODate2SysDate($StopDate);
		print "DEBUG: Comparing ReservedDomain $AdID todayfullsysdate = $todayfullsysdate, StopDate ==> $StopDate, " if $debugging == 1;
		if ( $todayfullsysdate >= $StopDate ) {
		    print "ad stopped on date. \n" if $debugging == 1;
		    next;
		} else {
		    print "ad still running. \n" if $debugging == 1;
		}
	    }
		    
	    $CurrentAdsList{$AdID} = 1;
		    
	    $reserved_domains_filedata .= "$domainname\t$position\t$AdID\n";
		    
	}
	close INFILE;
	
    } # if ad_reserved_domains_alloc.dat


# Save the lean mean ad_info_lite.dat file listing only current ads for this siteid

    $ad_info_data = '';
    foreach $AdID ( sort keys %CurrentAdsList ) {
	    $ad_info_data .= $AllAdsList{$AdID};
    }

    if ( $ad_info_data =~ /\w+/ ) {
	foreach $tmp_siteid ( @ad_allocation_subsites ) {
	    if ( $dontsavelogs == 1 ) {
		print "NOTICE: Not saving $datfilepath/$adinfolite.dat\.$tmp_siteid : dontsavelogs == 1 \n";
	    } else {  
		if ( open (OUTFILE, "> $datfilepath/$adinfolite\.$tmp_siteid" ) ) {
		    print OUTFILE $ad_info_data;
		    close OUTFILE;
		    print "DEBUG: Created new $datfilepath/$adinfolite\.$tmp_siteid file : " if $debugging == 1;
		    print length( $ad_info_data ) . " bytes saved. \n" if $debugging == 1;
		} else {
		    print "ERROR: Can't create $datfilepath/$adinfolite\.$tmp_siteid : $! \n";
		}
	    }
	} # foreach ad_allocation_subsite
    } else {
	print "ERROR: $adinfolite for $siteid came out empty: ad_info.dat may either be missing or empty. \n";
    }



# Save the new ad_reservations.dat file

    # if ( $reserved_filedata ) {

	foreach $tmp_siteid ( @ad_allocation_subsites ) { 
	    if ( $dontsavelogs == 1 ) {
		print "NOTICE: Not saving $datfilepath/ad_reservations.dat\.$tmp_siteid : dontsavelogs == 1 \n";
	    } else { 
		if ( open( OUTFILE, "> $datfilepath/ad_reservations.dat\.$tmp_siteid") ) {
		    print OUTFILE $reserved_filedata;
		    close OUTFILE;
		    print "DEBUG: Created new $datfilepath/ad_reservations.dat\.$tmp_siteid file : " . length($reserved_filedata) . " bytes \n" if $debugging == 1;
		} else {
		    print "ERROR: Could not recreate $datfilepath/ad_reservations.dat\.$tmp_siteid : $! \n";
		}
	    }
	} # foreach ad_allocation_subsite

    # } else {
    #	print "DEBUG: No reserved pages data to be saved for $siteid.\n" if $debugging == 1;
    # }

# Save the new ad_reserved_domains.dat file

    # if ( $reserved_domains_filedata ) {

	foreach $tmp_siteid ( @ad_allocation_subsites ) { 
	    if ( $dontsavelogs == 1 ) {
		print "NOTICE: Not saving $datfilepath/ad_reserved_domains.dat\.$tmp_siteid : dontsavelogs == 1 \n";
	    } else { 
		if ( open( OUTFILE, "> $datfilepath/ad_reserved_domains.dat\.$tmp_siteid") ) {
		    print OUTFILE $reserved_domains_filedata;
		    close OUTFILE;
		    print "DEBUG: Created new $datfilepath/ad_reserved_domains.dat\.$tmp_siteid file. : " . length($reserved_domains_filedata) . " bytes \n" if $debugging == 1;
		} else {
		    print "ERROR: Could not recreate $datfilepath/ad_reserved_domains.dat\.$tmp_siteid : $! \n";
		}
	    }
	} # foreach ad_allocation_subsite

    # } else {
    #	print "DEBUG: No reserved domains data to be saved for $siteid.\n" if $debugging == 1;
    # }

} # end sub setAdRotations





# Make a list of all ads in the system: Key each line in ad_info.dat table to its AdID

sub readAdInfo {

	if ( open (INFILE, "< $datfilepath/ad_info.dat" ) ) {
	    while (<INFILE>) {
		next if $_ !~ /^(\w+)\t/;
		$AllAdsList{$1} = $_;
	    }
	    close INFILE;
	    print "DEBUG: Read current AdID list from $datfilepath/ad_info.dat : " . (keys %AllAdsList) . " AdIDs listed.\n" if $debugging == 1;
	} else {
	    print "ERROR: Couldn't read $datfilepath/ad_info.dat: $! \n";
	    print "Won't be able to make new ad_info2.dat files for each site. \n";
	}

} # sub readAdInfo





# Make consolidated array of impressions and clicks for this ad on this site so far today

sub readAdStats {

    %AdData = ();
    %AdImpressions = ();

    if ( open(INFILE, "< $adlogfile") ) {

	while (<INFILE>) {
	    chomp;
	    
	    ($AdID, $impressions, $clicks, $tmp_siteid) = split(/\t/, $_);
# TODO: Can't decide whether to consider ad_allocation_group or ad_stats_group when considering impressions for StopOnTarget
	    $tmp_siteid = $sitesconfig{ $tmp_siteid . 'ad_allocation_group' } . '000en' if $sitesconfig{ $tmp_siteid . 'ad_allocation_group' };
	    $AdImpressions{ $tmp_siteid . $AdID } += $impressions;
		    
	}
	close INFILE;

	if ( $debugging == 1 ) {
	    $number_of_adloglines = (keys %AdImpressions);
	    print "DEBUG: readAdStats consolidated $number_of_adloglines lines from $adlogfile.\n";
	}

    } else {	    
	print "ERROR: readAdStats can't read $adlogfile: $! : This only affects StopOnTarget ads. \n";
	return;
    }

 }  # sub readAdStats





# Given an ad_x_alloc.dat file and siteid, create the corresponding ad_x_rotate.dat file

sub makeRotateFile {

    my $allocfile         = "$datfilepath/ad\_$ad_position\_alloc.dat\.$siteid";
    my $rotatefile        = '';
    my $lowest_allocation = 99999999;
    my %AdList            = ();
    my @results_list      = ();
    my $results           = '';
    my $results2          = '';
	

# Read current ad allocation for this ad position


	print "DEBUG: Reading $allocfile for $siteid...\n" if $debugging == 1;

	if ( open(INFILE, "< $allocfile") ) {
	    while (<INFILE>) {
		chomp;
		($AdID, $allocation, $stop_on_target, $StartDate, $StopDate) = split(/\t/, $_);
		# Ignore blank AdIds
		next if $AdID !~ /\w{10}/;
		# Ignore garbage entries
		next if $allocation < 10;
		# StopOnTarget (stop if daily number of impressions met or exceeded)
		if ( $stop_on_target ) {
		    $impressions_delivered = $AdImpressions{ $siteid . $AdID };
		    if ( $debugging == 1 ) {
			print "DEBUG: $siteid $AdID StopOnTarget, ";
			print "allocation = $allocation, impressions_delivered = $impressions_delivered : ";
		    }
		    if ( $impressions_delivered >= $allocation ) {
			print "ad stopped.\n" if $debugging == 1;
			next;
		    } else {
			print "ad allocated again.\n" if $debugging == 1;
		    }
		}
		# Consider the StartDate and StopDate when reallocating ads

		if ( $StartDate ) {
		    $StartDate = &ISODate2SysDate($StartDate);
		    print "DEBUG: $siteid $AdID StartDate todayfullsysdate = $todayfullsysdate, StartDate ==> $StartDate : " if $debugging == 1;
		    if ( $todayfullsysdate < $StartDate ) {
			print "ad not started yet. \n" if $debugging == 1;
			next;
		    } else {
			print "ad has started running. \n" if $debugging == 1;
		    }
		}

		if ( $StopDate ) {
		    $StopDate = &ISODate2SysDate($StopDate);
		    print "DEBUG: $siteid $AdID StopDate todayfullsysdate = $todayfullsysdate, StopDate ==> $StopDate, " if $debugging == 1;
		    if ( $todayfullsysdate >= $StopDate ) {
			print "ad stopped on date. \n" if $debugging == 1;
			next;
		    } else {
			print "ad still running. \n" if $debugging == 1;
		    }
		}

		print "DEBUG: $siteid $AdID allocated for $allocation impressions. " if $debugging == 1;


# TODO: Should the rotation formula be ( $allocation = $allocation - $impressions_delivered )
#       so that each ad is balanced according how many more impressions it needs???
#
#       Problem: allocation_remaining may be so low that it unabalances rotation list repeat factor (compare to largest allocation???)
#       Problem: Need to consider StopOnTarget and not let ads burnout ( re-run allocation if rotation is empty ???)

		if ( $testing_rebalanced_allocation == 1 ) {
		    $impressions_delivered = $AdImpressions{ $siteid . $AdID };
		    $allocation_remaining = $allocation - $impressions_delivered;
		    if ( $allocation_remaining > 10 ) {
			$allocation = $allocation_remaining;
			print " Delivered $impressions_delivered, allocating $allocation_remaining more.\n" if $debugging == 1;
		    }
		} else {
		    print "\n" if $debugging == 1;
		}

		$AdList{$AdID} = $allocation;
		$results2 .= "$AdID\n";
		$lowest_allocation = $allocation if $allocation < $lowest_allocation;
	    }
	    close INFILE;
	} else {
	    print "ERROR: Couldn't read $allocfile: $! \n";
	    return;
	}
	

	print "DEBUG: Lowest allocation in $allocfile is $lowest_allocation\n" if $debugging == 1;


# Create the new rotation list for this ad position

	foreach $AdID (keys %AdList) {
		$repeat = int( $AdList{$AdID} / $lowest_allocation );
		print "DEBUG: $AdID repeat factor = $repeat\n" if $debugging == 1 ;
		push(@results_list, $AdID) while ($repeat-- > 0);
		$CurrentAdsList{$AdID} = 1;
	}


	$results = join("\n", (&ShuffleList(@results_list)) );
	$results .= "\n";


# Save new rotation list file

	foreach $tmp_siteid ( @ad_allocation_subsites ) {

	    $rotatefile = "$datfilepath/ad\_$ad_position\_rotate.dat\.$tmp_siteid";

	    if ( $dontsavelogs == 1 ) {
		print "NOTICE: Not saving $rotatefile : dontsavelogs == 1 \n";
	    } else { 
		if ( open OUTFILE, "> $rotatefile" ) {
		    print OUTFILE $results;
		    print OUTFILE $results2 if ($results !~ /\w+/);
		    close OUTFILE;
		    if ( $debugging == 1 ) {
			$rotatefile_size = length $results;
			print "DEBUG: Created new $rotatefile : $rotatefile_size bytes saved.\n";
		    }
		} else {
		    print "ERROR: Could not create $rotatefile: $! \n";
		}
	    }

	} # foreach ad_allocation_subsite
	

 } # sub MakeRotateFile




sub ISODate2SysDate {

        local ($ISODate) = @_;
        if ($ISODate =~ /(\d+)\/(\w+)\/(\d+)/) {
                %monthlist = qw/Jan 01 Feb 02 Mar 03 Apr 04 May 05 Jun 06 Jul 07 Aug 08 Sep 09 Oct 10 Nov 11 Dec 12/;
                return $3 . $monthlist{$2} . $1;
        } else {
                return 0;
        } 

}  # Converts 02/Jan/1998 to 19980102 to do date comparisons





sub ShuffleList {

    my @result = @_;
    my $num_items = @result;
    while ($num_items > 1) {
	my $rand_item = rand $num_items;
	$num_items--;
	@result[$rand_item, $num_items] = @result[$num_items, $rand_item];
    }
    
    @result;

}  # sub ShuffleList randomly shuffles a list, straight out of "Effective Perl Programming", page 100 (item 47)




__END__
