
# Library of functions that produce graphs of a given stats hash
#
# Modified:   Kurt Gray -- Mon Jan 29 11:54:23 EST 2001     First rev.
#

use GD;


# sub makeGraph: given a global hash of stats, make a bar graph
#
# Inputs:        %graph_data    (global)    Hash of stats to graph:
#    
#                             X-Axis       Y-BarHeight   BarColor
#                             ------       -----------   --------
#                $graph_data{'20010214'}->{'1234567'} = 'Slashdot'
#                $graph_data{'20010214'}->{'234567'}  = 'FreshMeat'
#                $graph_data{'20010214'}->{'123566'}  = 'MediaBuilder'
#                $graph_data{'20010213'}->{'1354578'} = 'Slashdot'
#                $graph_data{'20010213'}->{'334535'}  = 'FreshMeat'
#                ...
#
#                $graph_title   (local)     Title text of the graph image
#
#
#
# Output:        $graph_url     (local)     URL of the GIF image of the output graph
#


sub makeGraph {
    
    my $graph_title  = shift;
    my $graph_width  = shift;
    my $graph_height = shift;

    # Set graph image colors:

    my %rgb = (
	       'white'  => [ 250, 250 ,250 ],
	       'black'  => [ 0,   0,   0   ],
	       'black2' => [ 70,  70,  70  ],
	       'gray'   => [ 200, 200, 200 ],
	       'red'    => [ 250, 0,   0   ],
	       'blue'   => [ 0,   0,   255 ],
	       'sand'   => [ 255, 255, 204 ],
	       'green'  => [ 0,   136, 0   ]
	       );
    my %color;

    my %barcolors = (
		     1 =>  [  65, 48,  130   ],
		     2 =>  [  52, 48,  182   ],
		     3 =>  [  138, 48,   48  ],
		     4 =>  [ 227, 65,   48   ],
		     5 =>  [ 227, 195,   48  ],
		     6 =>  [ 117, 117,   65  ],
		     7 =>  [  48,  182,   48 ],
		     8 =>  [ 203,  48,   48  ],
		     9 =>  [ 255, 255,   0     ],
		     10 => [ 0,   255,   255   ],
		     11 => [ 255,   0,  255   ],
		     );


    # Set graph height, width, and border margins:

    $graph_width   ||= 640;
    $graph_height  ||= 200;
    ( $margin_left, $margin_right ) = ( 10, 50 );
    ( $margin_top, $margin_bottom ) = ( 20, 70 );


    my $graph_image = new GD::Image($graph_width, $graph_height);
    while (my($k,$v) = each %rgb) { $color{$k} = $graph_image->colorAllocate($v->[0], $v->[1], $v->[2]) }; # set hash of colors


    # Draw graph image margins and borders:

    $graph_image->rectangle(0, 0 => $graph_width - 1, $graph_height-1, $color{'black'});        # black frame around gif
    $graph_image->fill(50, 50, $color{'white'});                                               # fill with white 
    $graph_image->rectangle(0+$margin_left, 0+$margin_top => 
			    $graph_width-$margin_right, $graph_height-$margin_bottom, $color{'gray'});   # area for the actual graph
    $graph_image->fill(5, 5, $color{'sand'});                                                  # fill area outside graph with light-yellow


    # Write graph title across top and center above graph:

    $graph_title_width = ( gdMediumBoldFont->width ) * length( $graph_title );
    $graph_image->string(gdMediumBoldFont, int(($graph_width - $graph_title_width) / 2), 5, $graph_title, $color{'black2'});



    # Determine the scales of the graph data:

    $numofbars       = ( keys %graph_data );

    # Stop here if graph data set is empty:
    if ( $numofbars < 1 ) {
	return 0;
    }

    $x_axis_length   = $graph_width - ( $margin_left + $margin_right );
    $y_axis_length   = $graph_height - ( $margin_top + $margin_bottom );
    $x_axis_height   = $graph_height - $margin_bottom;
    $bar_width       = $x_axis_length / $numofbars;
    $MAX_height      = 0;
    $MIN_height      = 9999999999;

    # Find the MAX bar height in the data set so we can scale all bars accordingly:

    foreach $tmp_date ( reverse sort keys %graph_data ) {

	$total_height = 0;

	foreach $tmp_name ( reverse sort { $graph_data{ $tmp_date }->{$a} <=> $graph_data{ $tmp_date }->{$b} } keys %{$graph_data{ $tmp_date }} ) {
	    $tmp_height = $graph_data{ $tmp_date }->{ $tmp_name };

	    $total_height += $tmp_height;

	    # Assign each bar a random color:

	    if ( ! $bar_colors{ $tmp_name } ) {

		    $bar_number++;

		    $tmp_color_R = $barcolors{ $bar_number % 12 }[0];
		    $tmp_color_G = $barcolors{ $bar_number % 12 }[1];
		    $tmp_color_B = $barcolors{ $bar_number % 12 }[2];

		    $bar_colors{ $tmp_name } ||= $graph_image->colorAllocate( $tmp_color_R, $tmp_color_G, $tmp_color_B );

		    # $graph_debug_data .= "$tmp_date: $tmp_name \t $tmp_height \t COLOR[$bar_number]: $tmp_color_R, $tmp_color_G, $tmp_color_B \n";
		    
		    # Draw the color keycode on left margin of graph:
		    
		    $color_key_x ||= $margin_left;
		    $color_key_y ||= $x_axis_height + 25;
		    $graph_image->filledRectangle( $color_key_x, $color_key_y - 3 => $color_key_x + 6, $color_key_y + 3, $bar_colors{ $tmp_name } );
		    $graph_image->string(gdSmallFont, $color_key_x + 10, $color_key_y - 6, $tmp_name, $bar_colors{ $tmp_name } );
		    $color_key_x += ( 20 * gdSmallFont->width ) if ($bar_number % 3) == 0;
		    $color_key_y = $x_axis_height + 25 + ( 15 * ($bar_number % 3));

	    } # if barcolor

	} # foreach height

	$MAX_height = $total_height if $MAX_height < $total_height;
	$MIN_height = $total_height if $MIN_height > $total_height;

    } # foreach date


    $height_scale = ( $y_axis_length / $MAX_height );



    # Mark MAX and MIN lines and label them on right margin:

    $MAX_axis_y = $graph_height - $margin_bottom - ( $height_scale * $MAX_height ) + 2;
    $graph_image->line( $margin_left, $MAX_axis_y  => $graph_width - $margin_right, $MAX_axis_y, $color{'gray'} );
    $graph_image->string(gdSmallFont, ( $graph_width - $margin_right + 2 ), $MAX_axis_y - 10, "MAX", $color{'black'} );
    $graph_image->string(gdSmallFont, ( $graph_width - $margin_right + 2 ), $MAX_axis_y, $MAX_height, $color{'black'} );
    $MIN_axis_y = $graph_height - $margin_bottom - ( $height_scale * $MIN_height ) + 2;
    $graph_image->line( $margin_left, $MIN_axis_y  => $graph_width - $margin_right, $MIN_axis_y, $color{'gray'} );
    $graph_image->string(gdSmallFont, ( $graph_width - $margin_right + 2 ), $MIN_axis_y - 10, "MIN", $color{'black'} );
    $graph_image->string(gdSmallFont, ( $graph_width - $margin_right + 2 ), $MIN_axis_y, $MIN_height, $color{'black'} );
   

    # Figure out where to mark dates along X axis:

    $x_label_width = ( gdSmallFont->width ) * 6;


    # Draw the bars on the graph:

    my ( $bar_x1, $bar_y1 ) = ( $margin_left, 0 );
    my ( $bar_x2, $bar_y2 ) = ( $margin_left + $bar_width, $x_axis_height );

    foreach $tmp_date ( sort keys %graph_data ) {

	# Mark of change of month and year on graph:

	$tmp_mon         = substr( $tmp_date, 4, 2 );
	$tmp_year        = substr( $tmp_date, 0, 4 );
	if ( $previous_mon && ( $tmp_mon ne $previous_mon )) {
	    $graph_image->dashedLine( $bar_x1, $margin_top => $bar_x1, $x_axis_height, $color{'gray'} );
	}
	if ( $previous_year && ( $tmp_year ne $previous_year )) {
	    $graph_image->line( $bar_x1, $margin_top => $bar_x1, $x_axis_height, $color{'gray'} );
	}
	$previous_mon = $tmp_mon;
	$previous_year = $tmp_year;


	# Draw each bar for this date:

	$previous_bar_y1 = $x_axis_height;

	foreach $tmp_name ( sort keys %bar_colors ) {
	    $tmp_height = $graph_data{ $tmp_date }->{ $tmp_name };
	    next if $tmp_height < 1;

	    # Draw this bar:
	    $bar_y2 = $previous_bar_y1;
	    $bar_y1 = $previous_bar_y1 - ( $tmp_height * ( $y_axis_length / $MAX_height ) );
	    $graph_image->filledRectangle( $bar_x1, $bar_y1 => $bar_x2, $bar_y2, $bar_colors{ $tmp_name });
	    $previous_bar_y1 = $bar_y1;

	    # $graph_debug_data .= "$tmp_date: $tmp_height: RECT: $bar_x1, $bar_y1 => $bar_x2, $bar_y2 \n";

	} # foreach bar

	# Label dates across x axis:

	$x_label_x1 = $bar_x1 + int(($bar_width - $x_label_width ) / 2);
	if ( $x_label_x1 > $previous_x_label ) {
	    $x_label_mark = $bar_x1 + int($bar_width / 2);
	    $graph_image->line( $x_label_mark, $x_axis_height, $x_label_mark, $x_axis_height + 4, $color{'black'} );
	    $graph_image->string(gdSmallFont, $x_label_x1, $x_axis_height + 3, &SysDate2MDate($tmp_date), $color{'black'} );
	    $previous_x_label = $x_label_x1 + $x_label_width;
	}
	# $graph_debug_data .= "$tmp_date: LABEL: \$x_label_x1 = $x_label_x1, \$previous_x_label = $previous_x_label \n";



	# Advance along the x axis:
 
	$bar_x1 += $bar_width;
	$bar_x2 += $bar_width;

    } # foreach tmp_date



    # Save the GIF file and return it's URL:

    $graph_filename = time . '.gif';
    $graph_filepath = "$graph_basepath/$graph_filename";

    if ( open FILE, "> $graph_filepath" ) {

	my $gif_data = $graph_image->gif;
	binmode FILE;
	print FILE $gif_data;
	close FILE;
    
        $graph_url   = "$graph_baseref/$graph_filename";
	return $graph_url;

    } else {
	print "ERROR: Could not save graph image $graph_filepath: $! \n";
	return 0;
    }


} # sub makeGraph



sub SysDate2MDate {

    my $SysDate = shift;
    if ( $SysDate =~ /^(\d{4})(\d{2})(\d{2})$/ ) {
	my ( $YYYY, $mm, $dd ) = ( $1, $2, $3 );
	my %monthlist = qw/01 Jan 02 Feb 03 Mar 04 Apr 05 May 06 Jun 07 Jul 08 Aug 09 Sep 10 Oct 11 Nov 12 Dec/;
	return "$monthlist{$mm}$dd";
    } else {
	return $SysDate;
    }

} # sub Sysdate2MDate



1;
