/*
 * mod_adbanners.c
 * version 1.05 - New Geotargeting; Megabanner w/o Navbar.dat fixed for NewsForge 
 *
 *  Copyright (C) 2000 Andover.Net/OSDN
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: kurt at osdn dot com
 *
 * Purpose in life:   This Apache module delivers HTML code for rotating ad banners in
 *                    separate simulataneous positions in each web page. The HTML code for each ad banner is
 *                    inserted into an environment variable which can be echoed in in a static HTML
 *                    web page using an SSI tag, or in a dynamic web page by echoing the contents
 *                    of the AD_BANNER environment variables.
 *
 *
 * How it works:      Each VirtualHost has their own set of text files that contain the
 *                    current ad banner rotation lists and HTML info for each ad. When Apache
 *                    starts (or each time a new httpd child is spawned) this module reads the ad data 
 *                    files into Apache's global memory heap ( in the cmd_parms->pool area ) then on each web
 *                    page request this module steps through each ad rotation banner list then 
 *                    creates new enviroment variables that hold the
 *                    the HTML code for the ad banners to be inserted in the web page via an
 *                    SSI tag like so:     <!--#echo var="AD_BANNER_1" -->  
 *
 *
 * Features:
 *                    - Client-side (JavaScript) cache-busting for more accurate ad impression counting
 *                    - Ability to reserve certain ads to certain pages in certain postions
 *                    - Ability to serve HTML layers or normal GIF banner ads
 *                    - Fault tolerant design so web server is not affected if ad banner files contain garbage
 *                    - ADBANNER_STATUS environment variable contains request handler flags useful for debugging
 *                    - MEGABANNERS: position 1 ad and position 2 can be synchronized to create wrap-around ads
 *
 *
 * httpd.conf Setup:  Each VirtualHost (web site) in httpd.conf that 
 *                    displays ad banners will need the following directives specified:
 *
 *                    Example mod_adbanners directives in httpd.conf:
 *
 *                    <VirtualHost www.acmeweb.com>
 *                    
 *                    AdBannerPagecountURL http://www.acmeweb.com/pagecount.gif		
 *                    AdBannerImagesBaseURL  /banner/
 *                    AdBannerHtmlAdDir /usr/local/apache/share/acmeweb/htdocs/htmlad/
 *
 *                    # IMPORTANT: AdBannerInfoFile directive must precede any AdBannerRotateFile directives!!!
 *                    #            Why, you ask? Because the ad_info.dat file has to be loaded before the
 *                    #            ad_x_rotate.dat files are processed.
 *
 *                    AdBannerInfoFile /usr/local/apache/share/acmeweb/cgi-bin/ad_info.dat
 *                    AdBannerRotateFile 1 /usr/local/apache/share/acmeweb/cgi-bin/ad_1_rotate.dat
 *                    AdBannerRotateFile 2 /usr/local/apache/share/acmeweb/cgi-bin/ad_2_rotate.dat
 *                    AdBannerRotateFile 3 /usr/local/apache/share/acmeweb/cgi-bin/ad_3_rotate.dat
 *                    AdBannerReservedPagesFile /usr/local/apache/share/acmeweb/cgi-bin/ad_reservations.dat
 *
 *                    # Right now mod_adbanners only supports up to 3 ad separate positions per web page 
 *
 *                    </VirtualHost>
 *
 * More Info:         Consult the mod_adbanners README file for more setup details.
 *
 * Notes:
 *                   - Apache rules
 *                   - Each http child process gets their own copy of module config record and data allocated in cmd->pool
 *                   - It seems I can't use rewind() on FILE pointers created by ap_pfopen() (Causes a segfault)
 *                   - It seems the mod_rewrite module is making "REDIRECT_" copies my ENV variables for some bizarre reason
 *                   - ap_psprintf() vs. sprintf(): ap_psprintf() assigns the address of destination char*, where sprintf() is given the location
 *                   - ad_psprintf() doesn't support %p format (handy for printing pointer addresses) but apparently does support %ul (u_long) 
 *                   - Using regexec() and ap_pregsub() in request handler function killed performance (Used %5 of CPU per request!)
 *
 *
 * Version history:
 *                   v1.00: 
 *                       - Only supported 3 simultaneous ad positions
 *                       - Required Apache to be restarted to read updated ad rotation lists
 *
 *                   v1.01:
 *                       - Added MEGABANNER support
 *                       - Installed on all of Andover.Net's web servers (expect Slashdot)
 *
 *                   v1.02:
 *                       - Support for 5 simultaneous ad positions
 *
 *                   v2.00 (forked by Marc Pilon):
 *                       - Support for unlimited number of simulatneous ad positions
 *                       - Uses shared memory + external update process so Apache did not have to be restarted
 *                       - Most strings buffered in restricted-length structures and arrays
 *                       - Installed on Slashdot at Exodus
 *
 *                   v1.03 (continuation of v1.02 code by Kurt Gray):
 *                       - Support for unlimited number of simulatneous ad positions
 *                       - ad rotation lists and ad_info table read during child_init() phase; does not require Apache restart
 *                       * WARNING * MaxRequestsPerChild must be set reasonably low (like around 300) so child_init() happens periodically
 *               
 *                   v1.04 (continuation of v1.03 code by Kurt Gray):
 *                       - Added basic support for geotargeting according to last two characters of ad_id
 *                       - Set IMG BORDER=0 and HREF TARGET=_top for better IFRAME/ILAYER compatibility
 *                       - Position 2 ad can not be overridden by MEGABANNER part B if position 2 is reserved
 *
 *                   v1.05 (Kurt Gray)
 *                       - Navbar's not required to run MEGABANNERs (NewsForge)
 *                       - Cleaned up reserved pages and geo-targeting logic
 *
 *
 * Modified:         Kurt Gray -- Wed Dec  1 11:12:04 EST 1999       Prints pagecount tag even if position 1 rotation is empty
 *                   Kurt Gray -- Fri May 12 18:18:04 EDT 2000       ad_info and ad_x_rotate files loaded on child_init()
 *                   Kurt Gray -- Tue May 16 16:21:33 EDT 2000       Moving ad_reservations read into child_init(); fixing multiple virthost bug
 *                   Kurt Gray -- Thu May 18 13:18:42 EDT 2000       Adding impression counter code to HTML ad layers
 *                   Kurt Gray -- Fri Jun  2 15:47:50 EDT 2000       Split up javascript doc.write() lines to fix FreeBSD+Netscape crash 
 *                   Kurt Gray -- Wed Oct 11 16:02:17 EDT 2000       Merged changes from ads.linux.com and sun00472
 *                   Kurt Gray -- Mon Oct 16 17:26:48 EDT 2000       Geotargeting according to last two chars in ad_id
 *                   Kurt Gray -- Fri Oct 27 20:49:09 EDT 2000       Fixed Reserved page + Geotargeting crash due to use of undefined variables
 *                   Kurt Gray -- Fri Apr 13 14:39:38 EDT 2001       Rewrote geotargeting to be more direct, requires HostnameLookups On
 *
 *
 * TODO:             See various TODO notes scattered throughout this file.
 *                   
 */


#include "httpd.h"
#include "http_config.h"
#include "http_log.h"	 /* required in order to call ap_log_error() function */

/* Various data size limits to prevent gratuitous memory consumption */

#define MAX_NAVBAR_FILESIZE    5048    /* Max number of bytes to read from a navbar.dat file */
#define MAX_ADINFO_FILESIZE    131072  /* Max number of bytes to read from an ad_info.dat file */
#define MAX_ROTATION_FILESIZE  64480   /* Max number of bytes to read from an ad_rotate file */
#define MAX_ROTATION_LISTSIZE  5024    /* Max number of intergers to store for an ad rotation list based on the MAX_ROTATION_FILESIZE */
#define MAX_LINE_LENGTH        4096    /* Max buffer size for reading file lines using fgets() */
#define MAX_ADID_OFFSET_SIZE   6       /* Size of string buffer that will translate adid offset from ascii to int */
#define MAX_PAGENAME_LENGTH    48      /* Max number of chars to read from a web page document name string */
#define MAX_AD_POSITIONS       6       /* Number of ad positions per page this module will serve, plus 1 */
#define ADID_SIZE              10      /* Number of characters long we expect in an Ad banner ID to be */

/* Compile-time flags */

#define ABDEBUG                0  /* Set to 1 to for status messages in error_log and during httpd startup */
#define ABDEBUG_VERBOSE        0  /* Set to 1 to see even more status messages in error_log and during httpd startup */
#define MEGABANNERS            1  /* Enable Andover.Net-style megabanners w/ navbar via AdBannerNavbar directive */
#define GEOTARGETING           1  /* Enable ads to be delivered targeted to last two chars user's remote_hostname */
#define IFRAME_SERVER          0  /* Determine pagename from HTTP_REFERER header */
#define PAGECOUNTGIF_BELOW_AD  0  /* Set to 1 for pagecount 1x1 pixel gif appear below the top ad (position 1) instead of above it */
#define PROCESS_ALL_REQUESTS   0  /* Set to 0 to deactivate this module on HTTP requests that are not .html, .pl, or .php web pages */
#define RANDOM_ADS             1  /* Set to 0 is sequential round-robin ad delivery only works for http children < StartServers setting */


/* 
   This typdef statement declares the format of an adbanners_server_rec record structure 
   for which each VirtualHost will have its own adbanners_server_rec to store global data
   in persistent memory during the entire lifetime of the parent server.

   These are basicially a list of things we need to know on a request-by-request basis 
   in order to create the full HTML code for the next ad bannar to be delivered.
 */


typedef struct {
  char *ad_pagecount_url;                          /* Pointer to a string that is the URL of pagecount.gif */
  char *ad_pagecount_url2;                         /* Pointer to a string that is the second URL of pagecount.gif */
  char *ad_click_url;                              /* Pointer to a string that is the URL of ad click redirector script */
  char *ad_images_url;	                           /* Pointer to a string that is the base URL for ad banner images */
  char *ad_htmlad_dir;	                           /* Pointer to a string that is the full path of the HTML ad source directory */
  char *ad_info_base;		                   /* Pointer to a string that is the entire ad_info table */
  char *ad_info_filepath;
  char *ad_x_rotate_filepath[MAX_AD_POSITIONS];    /* Array of strings containing full paths to each ad rotation list file */
  int  *ad_x_rotate[MAX_AD_POSITIONS];		   /* Array of pointers to the base of the position x ad rotation list in memory */
  int   ad_x_numberofads[MAX_AD_POSITIONS];	   /* Array of integers numbers of ads rotating in position x */
  int   ad_x_currentad[MAX_AD_POSITIONS];	   /* Array of list indexes of the next ad to be delivered in position x */
  char *ad_reserved_pages_filepath;
  char *ad_reserved_pages;	                   /* Pointer to a string that is contents of the reserved pages table */

#if GEOTARGETING
  char *ad_reserved_domains_filepath;
  char *ad_reserved_domains;	                   /* Pointer to a string that is contents of the reserved domains table */
#endif

#if MEGABANNERS
  char *ad_navbar_html;                            /* A string that is the contents of the local navbar.txt and bgcolor.dat file */
  char *ad_navbar_gifsurl;                         /* A string containing the URL of the decorative GIFs server */
  char *ad_navbar_bgcolor;                         /* A string containting the page background color code for this web site */
#endif
} adbanners_server_rec;




/*
   Declaring the name of this Apache module to the Apache compiler
 */

module MODULE_VAR_EXPORT adbanners_module;



/*
   This function allocates an adbanners_server_rec record on the global 
   server memory heap for each VirtualHost. It should be called whenever Apache 
   is started/restarted, and that occures because we listed this function as the
   server config handler in the MODULE_EXPORT table at the bottom of this file.
 */


static void *create_adbanners_server_config(pool *p, server_rec *dummy)
{

  int i;

  adbanners_server_rec *new = (adbanners_server_rec *) ap_palloc(p, sizeof(adbanners_server_rec));

  new->ad_pagecount_url = NULL;
  new->ad_pagecount_url2 = NULL;
  new->ad_click_url = NULL;
  new->ad_images_url = NULL;
  new->ad_htmlad_dir = NULL;
  new->ad_info_base = NULL;
  new->ad_info_filepath = NULL;
  for (i=0; i < MAX_AD_POSITIONS; i++) {
    new->ad_x_rotate_filepath[i] = NULL;
    new->ad_x_rotate[i] = NULL;
    new->ad_x_numberofads[i] = 0;
    new->ad_x_currentad[i] = 0;
  }
  new->ad_reserved_pages_filepath = NULL;
  new->ad_reserved_pages = NULL;

#if GEOTARGETING
  new->ad_reserved_domains_filepath = NULL;
  new->ad_reserved_domains = NULL;
#endif

#if MEGABANNERS
  new->ad_navbar_html = NULL;
  new->ad_navbar_gifsurl = NULL;
  new->ad_navbar_bgcolor = NULL;
#endif

    return (void *) new;
}



/* 
   merge_adbanners_server_config() overrides parent server directives in httpd.conf 
   with VirtualHost settings
 */


static void *merge_adbanners_server_config(pool *p, void *base, void *new)
{

  int i;

  adbanners_server_rec *merged = (adbanners_server_rec *) ap_palloc(p, sizeof(adbanners_server_rec));
  adbanners_server_rec *parent = (adbanners_server_rec *) base;
  adbanners_server_rec *child = (adbanners_server_rec *) new;

  merged->ad_pagecount_url = child->ad_pagecount_url ? child->ad_pagecount_url : parent->ad_pagecount_url;
  merged->ad_pagecount_url2 = child->ad_pagecount_url2 ? child->ad_pagecount_url2 : parent->ad_pagecount_url2;
  merged->ad_click_url = child->ad_click_url ? child->ad_click_url : parent->ad_click_url;
  merged->ad_images_url = child->ad_images_url ? child->ad_images_url : parent->ad_images_url;
  merged->ad_htmlad_dir = child->ad_htmlad_dir ? child->ad_htmlad_dir : parent->ad_htmlad_dir;
  merged->ad_info_base = child->ad_info_base ? child->ad_info_base : parent->ad_info_base;
  merged->ad_info_filepath = child->ad_info_filepath ? child->ad_info_filepath : parent->ad_info_filepath;
  merged->ad_reserved_pages_filepath = child->ad_reserved_pages_filepath ? child->ad_reserved_pages_filepath : parent->ad_reserved_pages_filepath;  
  merged->ad_reserved_pages = child->ad_reserved_pages ? child->ad_reserved_pages : parent->ad_reserved_pages;

  for (i=0; i < MAX_AD_POSITIONS; i++) {
    merged->ad_x_rotate_filepath[i] = child->ad_x_rotate_filepath[i] ? child->ad_x_rotate_filepath[i] : parent->ad_x_rotate_filepath[i];
    merged->ad_x_rotate[i] = child->ad_x_rotate[i] ? child->ad_x_rotate[i] : parent->ad_x_rotate[i];
    merged->ad_x_numberofads[i] = child->ad_x_numberofads[i] ? child->ad_x_numberofads[i] : parent->ad_x_numberofads[i];
    merged->ad_x_currentad[i] = child->ad_x_currentad[i] ? child->ad_x_currentad[i] : parent->ad_x_currentad[i];
  }

#if GEOTARGETING
  merged->ad_reserved_domains_filepath = child->ad_reserved_domains_filepath ? child->ad_reserved_domains_filepath : parent->ad_reserved_domains_filepath;
  merged->ad_reserved_domains = child->ad_reserved_domains ? child->ad_reserved_domains : parent->ad_reserved_domains;
#endif

#if MEGABANNERS
  merged->ad_navbar_html = child->ad_navbar_html ? child->ad_navbar_html : parent->ad_navbar_html;
  merged->ad_navbar_gifsurl = child->ad_navbar_gifsurl ? child->ad_navbar_gifsurl : parent->ad_navbar_gifsurl;
  merged->ad_navbar_bgcolor = child->ad_navbar_bgcolor ? child->ad_navbar_bgcolor : parent->ad_navbar_bgcolor;
#endif

    return (void *) merged;

} /* merge_server_config */







/*

   The following functions are handlers for each different type of server directives
   in the httpd.conf file that pertains to our own module. Our input to these functions
   comes from specifying the TAKE1 flag in the adbanners_module_cmds[] array.

 */







/*
   This function handles the AdBannerPagecountURL directive in the httpd.conf file
   This is the primary pagecount server URL. The AdBannerPagecountBackupURL directive
   is available to specify a secondary pagecount server in case the primary goes down.
 */

static const char *adbanners_module_read_config_pagecounturl(cmd_parms *cmd, void *mconfig, const char *arg)
{

    adbanners_server_rec *sconf = NULL;		/* A pointer to our module data record */
    char *url = NULL;	                        /* Pointer to our reserved memory area to hold sconf->ad_pagecount_url */


    /* 
       TODO: Warn user if images URL is invalid:
       We can't verify the images directory exists unless it's a relative (local) URL so if the URL starts with a '/'
       then check to make sure the directory exists locally. If it's a relative URL but the directory
       is missing then log the error but continue as usual using the erroneous URL anyway and hopefully 
       the webmaster will notice he's shooting blanks.
     */



    /* Get a pointer to our server config record... */

    sconf = (adbanners_server_rec *) ap_get_module_config(cmd->server->module_config, &adbanners_module);

    /* Allocate space for our pagecount_url string in the server heap */

    if ((url = ap_pcalloc(cmd->pool, (strlen(arg) + 1) * sizeof(char))) == NULL) {
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     cmd->server,
		     "%s: AdBannerPagecountURL error: Apache would not allocate %d bytes of memory to store url string.",
		     __FILE__, 
		     (strlen(arg) + 1));
	return NULL;
    }

    sprintf(url, arg);

    /* Insert pointer to the images_url string in the server config record */

    sconf->ad_pagecount_url = url;

#if ABDEBUG
    printf("%s AdBannerPagecountURL = \"%s\"\n", cmd->server->server_hostname, sconf->ad_pagecount_url);
#endif


    return NULL;


} /* read_config_pagecounturl */





/*
   This function handles the AdBannerPagecountBackupURL directive in the httpd.conf file
   This is a secondary pagecount URL in case the primary pagecount server is down.
 */

static const char *adbanners_module_read_config_pagecountbackupurl(cmd_parms *cmd, void *mconfig, const char *arg)
{

    adbanners_server_rec *sconf = NULL;		/* A pointer to our module data record */
    char *url = NULL;	                        /* Pointer to our reserved memory area to hold sconf->ad_pagecount_url */


    /* 
       TODO: Warn user if images URL is invalid:
       We can't verify the images directory exists unless it's a relative (local) URL so if the URL starts with a '/'
       then check to make sure the directory exists locally. If it's a relative URL but the directory
       is missing then log the error but continue as usual using the erroneous URL anyway and hopefully 
       the webmaster will notice he's shooting blanks.
     */



    /* Get a pointer to our server config record... */

    sconf = (adbanners_server_rec *) ap_get_module_config(cmd->server->module_config, &adbanners_module);

    /* Allocate space for our pagecount_url string in the server heap */

    if ((url = ap_pcalloc(cmd->pool, (strlen(arg) + 1) * sizeof(char))) == NULL) {
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     cmd->server,
		     "%s: AdBannerPagecountBackupURL error: Apache would not allocate %d bytes of memory to store url string.",
		     __FILE__, 
		     (strlen(arg) + 1));
	return NULL;
    }

    sprintf(url, arg);

    /* Insert pointer to the images_url string in the server config record */

    sconf->ad_pagecount_url2 = url;

#if ABDEBUG
    printf("%s AdBannerPagecountBackupURL = \"%s\"\n", cmd->server->server_hostname, sconf->ad_pagecount_url2);
#endif


    return NULL;


} /* read_config_pagecountbakcupurl */







/*
   This function handles the AdBannerClickURL directive in the httpd.conf file
   The argumaent is the full or relative URL to a script that will redirect banner clicks.
 */

static const char *adbanners_module_read_config_clickurl(cmd_parms *cmd, void *mconfig, const char *arg)
{

    adbanners_server_rec *sconf = NULL;		/* A pointer to our module data record */
    char *url = NULL;	                        /* Pointer to our reserved memory area to hold sconf->ad_pagecount_url */


    /* 
       TODO: Warn user if images URL is invalid:
       We can't verify the images directory exists unless it's a relative (local) URL so if the URL starts with a '/'
       then check to make sure the directory exists locally. If it's a relative URL but the directory
       is missing then log the error but continue as usual using the erroneous URL anyway and hopefully 
       the webmaster will notice he's shooting blanks.
     */



    /* Get a pointer to our server config record... */

    sconf = (adbanners_server_rec *) ap_get_module_config(cmd->server->module_config, &adbanners_module);

    /* Allocate space for our pagecount_url string in the server heap */

    if ((url = ap_pcalloc(cmd->pool, (strlen(arg) + 1) * sizeof(char))) == NULL) {
	ap_log_error(APLOG_MARK, APLOG_NOERRNO | APLOG_NOTICE, cmd->server,
		     "%s: AdBannerClickURL error: Apache would not allocate %d bytes of memory to store url string.",
		     __FILE__, (strlen(arg) + 1));
	return NULL;
    }

    sprintf(url, arg);

    /* Insert pointer to the images_url string in the server config record */

    sconf->ad_click_url = url;

#if ABDEBUG
    printf("%s AdBannerClickURL = \"%s\"\n", cmd->server->server_hostname, sconf->ad_click_url);
#endif


    return NULL;


} /* read_config_clickurl */









/*
   This function handles the AdBannerImagesBaseURL directive in the httpd.conf file
   The argument is the full or relative URL too precede the IMG tags for each banner impression.
 */

static const char *adbanners_module_read_config_imagesurl(cmd_parms *cmd, void *mconfig, const char *arg)
{

    adbanners_server_rec *sconf = NULL;		/* A pointer to our module data record */
    char *url = NULL;	                        /* Pointer to our reserved memory area to hold sconf->ad_images_url */


    /* 
       TODO: Warn user if images URL is invalid:
       We can't verify the images directory exists unless it's a relative (local) URL so if the URL starts with a '/'
       then check to make sure the directory exists locally. If it's a relative URL but the directory
       is missing then log the error but continue as usual using the erroneous URL anyway and hopefully 
       the webmaster will notice he's shooting blanks.
     */



    /* Get a pointer to our server config record... */

    sconf = (adbanners_server_rec *) ap_get_module_config(cmd->server->module_config, &adbanners_module);

    /* Allocate space for our images_url string in the server heap */

    if ((url = ap_pcalloc(cmd->pool, (strlen(arg) + 1) * sizeof(char))) == NULL) {
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     cmd->server,
		     "%s: AdBannerImagesBaseURL error: Apache would not allocate %d bytes of memory to store url string.",
		     __FILE__, 
		     (strlen(arg) + 1));
	return NULL;
    }

    sprintf(url, arg);

    /* Insert pointer to the images_url string in the server config record */

    sconf->ad_images_url = url;

#if ABDEBUG
    printf("%s AdBannerImagesBaseURL = \"%s\"\n", cmd->server->server_hostname, sconf->ad_images_url);
#endif


    return NULL;


} /* read_config_imagesurl */









/*
   This function handles the AdBannerHtmlAdDir directive in the httpd.conf file
   The argument is the relative path on the local machine from which to read layer files for HTML ads.
 */


static const char *adbanners_module_read_config_htmladdir(cmd_parms *cmd, void *mconfig, const char *arg)
{

    adbanners_server_rec *sconf = NULL;		/* A pointer to our module data record */
    char *ad_htmladdir = NULL;	/* Pointer to our reserved memory area to hold sconf->ad_htmlad_dir */


    /* Abort this function if directory is invalid */

    if (ap_is_directory(arg) < 1) {
	ap_log_error(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, cmd->server,
		     "%s: Can't read directory \"%s\" given in AdBannerHtmlAdDir directive: Must be a full pathname to a local directory.",
		     __FILE__, arg);
	return NULL;
    }

    /* Get a pointer to our server config record... */

    sconf = (adbanners_server_rec *) ap_get_module_config(cmd->server->module_config, &adbanners_module);

    /* Allocate space for our images_url string in the server heap */

    if ((ad_htmladdir = ap_pcalloc(cmd->pool, (strlen(arg) + 1) * sizeof(char))) == NULL) {
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     cmd->server,
		     "%s: AdBannerHtmlAdDir error: Apache would not allocate %d bytes of memory to store ad_htmlad_dir.",
		     __FILE__, 
		     (strlen(arg) + 1));
	return NULL;
    }

    sprintf(ad_htmladdir, arg);

    /* Insert pointer to the images_url string in the server config record */

    sconf->ad_htmlad_dir = ad_htmladdir;

#if ABDEBUG
    printf("%s AdBannerHtmlAdDir = \"%s\"\n", cmd->server->server_hostname, ad_htmladdir);
#endif

    return NULL;

} /* read_config_htmladdir */










/*
   This function handles the AdBannerInfoFile directive in the httpd.conf file
   The argument is the full path on local machine from which to read the ad_info.dat file.
 */


static const char *adbanners_module_read_config_adinfofile(cmd_parms *cmd, void *mconfig, const char *filepath)
{

    adbanners_server_rec *sconf = NULL;		/* A pointer to our module data record */
    char *adinfo_filepath = NULL;	        /* pointer to a string that is the path to the ad_info text file */


    /* Get a pointer to our server config record... */

    sconf = (adbanners_server_rec *) ap_get_module_config(cmd->server->module_config, &adbanners_module);

    /* Allocate space for our adinfo_filepath string in the server heap */

    if ((adinfo_filepath = ap_pcalloc(cmd->pool, (strlen(filepath) + 1) * sizeof(char))) == NULL) {
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_ERR, 
		     cmd->server,
		     "%s: AdBannerInfoFile error: Apache would not allocate %d bytes of memory to store adinfo_filepath string.",
		     __FILE__, 
		     (strlen(filepath) + 1));
	return NULL;
    }

    sprintf(adinfo_filepath, filepath);

    sconf->ad_info_filepath = adinfo_filepath;


#if ABDEBUG
    printf("%s AdBannerInfoFile = \"%s\"\n", cmd->server->server_hostname, adinfo_filepath);
#endif


    return NULL;

} /* read_config_adinfofile */








/*
   This function handles the AdBannerRotateFile directive in the httpd.conf file
   The arguments are the ad position number then the full path to read the rotation list file for that ad position.
 */


static const char *adbanners_module_read_config_rotatelist(cmd_parms *cmd, void *mconfig, const char *arg1, const char *filepath)
{

    adbanners_server_rec *sconf = NULL;		   /* A pointer to our module data record */
    int   ad_position = 0;	                   /* integer translation of arg1 AdBannerRotateFile argument */
    char *rotate_filepath = NULL;                  /* A string that is the path of ad rotation file */


    /* arg1 will the ad position number and arg2 will be the full path to the rotate file */

    ad_position = atoi(arg1);


    /* Check the given position number argument is a number between 1 and 3, inclusive */
    /* If not given a valid ad position number then log the error and get out of this function */

    if ( ( ad_position < 1 ) || ( ad_position > MAX_AD_POSITIONS ) ) {
	ap_log_error(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, cmd->server,
		     "%s: \"%s\" is not a valid ad position number for the AdBannerRotateFile directive. Try a number 1 through %d instead.",
		     __FILE__, arg1, MAX_AD_POSITIONS);
	return NULL;
    }


    /* Get a pointer to our server config record... */

    sconf = (adbanners_server_rec *) ap_get_module_config(cmd->server->module_config, &adbanners_module);



    /* Allocate space for our rotate_filepath string in the server heap */

    if ((rotate_filepath = ap_pcalloc(cmd->pool, (strlen(filepath) + 1) * sizeof(char))) == NULL) {
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_ERR, 
		     cmd->server,
		     "%s: AdBannerRotateFile error: Apache would not allocate %d bytes of memory to store rotate_filepath string.",
		     __FILE__, 
		     (strlen(filepath) + 1));
	return NULL;
    }

    sprintf(rotate_filepath, filepath);


    sconf->ad_x_rotate_filepath[ad_position] = rotate_filepath;

#if ABDEBUG
    printf("%s AdBannerRotateFile[%d] = \"%s\"\n", cmd->server->server_hostname, ad_position, rotate_filepath);
#endif

    return NULL;

}  /* end of adbanners_module_read_config_rotatelist */










/*
   This function handles the AdBannerReservedPagesFile directive in the httpd.conf file 
 */


static const char *adbanners_module_read_config_reservedlist(cmd_parms *cmd, void *mconfig, const char *filepath)
{

    adbanners_server_rec *sconf = NULL;		/* A pointer to our module data record */
    char *adreserved_filepath = NULL;	        /* pointer to a string that is the path to reserved pages file */


    /* Get a pointer to our server config record... */

    sconf = (adbanners_server_rec *) ap_get_module_config(cmd->server->module_config, &adbanners_module);

    /* Allocate space for our adinfo_filepath string in the server heap */

    if ((adreserved_filepath = ap_pcalloc(cmd->pool, (strlen(filepath) + 1) * sizeof(char))) == NULL) {
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_ERR, 
		     cmd->server,
		     "%s: AdBannerReservedPagesFile error: Apache would not allocate %d bytes of memory to store adreserved_filepath string.",
		     __FILE__, 
		     (strlen(filepath) + 1));
	return NULL;
    }

    sprintf(adreserved_filepath, filepath);

    sconf->ad_reserved_pages_filepath = adreserved_filepath;

#if ABDEBUG
    printf("%s AdBannerReservedPagesFile = \"%s\"\n", cmd->server->server_hostname, adreserved_filepath);
#endif

    return NULL;

}  /* end of config_reservedlist handler */






#if GEOTARGETING


/*
   This function handles the AdBannerReservedDomainsFile directive in the httpd.conf file 
 */


static const char *adbanners_module_read_config_reserveddomains_list(cmd_parms *cmd, void *mconfig, const char *filepath)
{

    adbanners_server_rec *sconf = NULL;		/* A pointer to our module data record */
    char *adreserved_filepath = NULL;	        /* pointer to a string that is the path to reserved pages file */


    /* Get a pointer to our server config record... */

    sconf = (adbanners_server_rec *) ap_get_module_config(cmd->server->module_config, &adbanners_module);

    /* Allocate space for our adinfo_filepath string in the server heap */

    if ((adreserved_filepath = ap_pcalloc(cmd->pool, (strlen(filepath) + 1) * sizeof(char))) == NULL) {
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_ERR, 
		     cmd->server,
		     "%s: AdBannerReservedDomainsFile error: Apache would not allocate %d bytes of memory to store adreserved_filepath string.",
		     __FILE__, 
		     (strlen(filepath) + 1));
	return NULL;
    }

    sprintf(adreserved_filepath, filepath);

    sconf->ad_reserved_domains_filepath = adreserved_filepath;

#if ABDEBUG
    printf("%s AdBannerReservedDomainsFile = \"%s\"\n", cmd->server->server_hostname, adreserved_filepath);
#endif

    return NULL;

}  /* end of config_reserveddomains_list handler */


#endif







#if MEGABANNERS


/*
   This function handles the AdBannerNavbar directive in the httpd.conf file.

   Example: AdBannerNavbar CCFFCC http://209.192.217.100/mb/ /usr/local/apache/sites/mediabuilder.com/htdocs/navbar.dat
 */


static const char *adbanners_module_read_config_navbarinfo(cmd_parms *cmd, void *mconfig, const char *colorcode, const char *gifserverurl, const char *filepath)
{

    struct stat finfo;		                /* A file stat structure to get size of file */
    int fd = 0;		                        /* A file descriptor of the file we are reading */
    adbanners_server_rec *sconf = NULL;		/* A pointer to our module data record */
    char *navbar_filecontents = NULL;	        /* A string that we read our file contents into */
    char *bg_color = NULL;
    char *gifs_url = NULL;
    int filesize = 0;	                        /* The raw size of the file we're about to read */


    /* Check to make sure the given path argument is a valid path to a file of reasonable size... */
    /* If that's not a valid file then log the error and leave the ad_1_rotate list pointer NULL so we ignore it */

    if (stat(filepath, &finfo) == -1) {

	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_ERR, 
		     cmd->server,
		     "%s: Can't read file \"%s\" given in AdBannerNavbar directive: %s",
		     __FILE__, 
		     filepath, 
		     strerror(errno));
	return NULL;

    } else {

	if (finfo.st_size > MAX_NAVBAR_FILESIZE) {

	    ap_log_error(APLOG_MARK, 
			 APLOG_NOERRNO | APLOG_NOTICE, 
			 cmd->server,
			 "%s: AdBannerNavbar file \"%s\" is too big (%ld bytes) so I'm only going to read the first %d bytes from it.", 
			 __FILE__, 
			 filepath, 
			 finfo.st_size, MAX_ADINFO_FILESIZE);

	    filesize = MAX_NAVBAR_FILESIZE;

	}

	filesize = finfo.st_size;

    } /* if stat filepath */


    /* Get contents of the navbar.dat file */

    if ((fd = ap_popenf(cmd->temp_pool, filepath, O_RDONLY, 0)) < 1) {

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_ERR, 
		   cmd->server, 
		   "%s: Can't open file \"%s\" given in AdBannerNavbar directive: %s",
		   __FILE__, 
		   filepath, 
		   strerror(errno));
      
    } else {

      /* Store contents of navbar.dat file in global server pool memory */

      if ((navbar_filecontents = ap_pcalloc(cmd->pool, (filesize + 1) * sizeof(char))) == NULL) {
	    ap_log_error(APLOG_MARK, 
			 APLOG_NOERRNO | APLOG_NOTICE, 
			 cmd->server,
			 "%s: Apache would not allocate %d bytes of memory to hold the [AdBannerNavbar] file.", 
			 __FILE__, 
			 filesize);
      } else {

	read(fd, navbar_filecontents, filesize);
	ap_pclosef(cmd->temp_pool, fd);

      }

    } /* if ap_popen */

    
    /* Store contents of gifserverurl string in global server pool memory */

    if ((gifs_url = ap_pcalloc(cmd->pool, ((strlen(gifserverurl)) + 1) * sizeof(char))) == NULL) {
      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   cmd->server,
		   "%s: Apache would not allocate %d bytes of memory to hold the [AdBannerNavbar] decorative GIF server url.", 
		   __FILE__, 
		   ((strlen(gifserverurl)) + 1) );
    } else {
      sprintf(gifs_url, "%s", gifserverurl);
    }



    /* Store contents of bgcolor string in global server pool memory */

    if ((bg_color = ap_pcalloc(cmd->pool, ((strlen(colorcode)) + 1) * sizeof(char))) == NULL) {
      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   cmd->server,
		   "%s: Apache would not allocate %d bytes of memory to hold the [AdBannerNavbar] bgcolor code.", 
		   __FILE__, 
		   ((strlen(colorcode)) + 1) );
    } else {
      sprintf(bg_color, "%s", colorcode);
    }


    /* Get a pointer to our server config record... */

    sconf = (adbanners_server_rec *) ap_get_module_config(cmd->server->module_config, &adbanners_module);


    /* Save the navbar parts in the server config record */

    sconf->ad_navbar_html = navbar_filecontents;
    sconf->ad_navbar_gifsurl = gifs_url;
    sconf->ad_navbar_bgcolor = bg_color;


#if ABDEBUG

    printf("%s MEGABANNERS are enabled. Processing AdBannerNavbar directive...\n", cmd->server->server_hostname);
    printf("%s AdBannerNavbar: sconf->ad_navbar_html = [%s]\n", cmd->server->server_hostname, filepath);
    printf("%s AdBannerNavbar: sconf->ad_navbar_gifsurl = \"%s\"\n", cmd->server->server_hostname, sconf->ad_navbar_gifsurl);
    printf("%s AdBannerNavbar: sconf->ad_navbar_bgcolor = \"%s\"\n\n", cmd->server->server_hostname, sconf->ad_navbar_bgcolor);

#endif


    return NULL;


}  /* If MEGABANNERS then process the AdBannerNavbar directive in httpd.conf */


#endif







/*

  adbanners_child_init() re-reads ad_x_rotate.dat, ad_info.dat, and ad_reservations.dat on httpd child_init.
  This is a new hack in order to avoid restarting Apache every hour to read new ad rotation lists. -Kurt

*/

void adbanners_child_init(server_rec *main_server, pool *p) {

  server_rec *server               = NULL;   /* pointer to each VirtualHost record in httpd.conf */
  adbanners_server_rec *sconf      = NULL;   /* pointer to the configuration record for this module */
  struct stat finfo;		             /* file stat structure to get size of file */
  FILE *fp                         = NULL;   /* pointer to the ad_x_rotate.dat we are reading */
  char  line_buffer[MAX_LINE_LENGTH];        /* String buffer to read each line of a file */
  char *adinfo                     = NULL;
  char *adinfo_base                = NULL;
  char *ad_info                    = NULL;   /* Copy of sconf->ad_info_base */
  char *ad_info_ptr                = NULL;   /* Index pointer to part of ad_info table */
  char *ad_id                      = NULL;   /* Ad ID read from ad_info_line */
  char *ad_info_line               = NULL;   /* Ad ID read from ad_info_line */
  int   indexer                    = 0;
  int   adinfo_size                = 0;      /* Size of ad_info.dat for to allocate memort block to hold it */
  int   listsize = 0, rawlistsize  = 0;	     /* Integers to calulate how mush space to allocate for rotation list */
  int  *rotation_list_base         = NULL;   /* The permanent memory address of the processed rotation list */
  int  *rotation_list_base_ptr     = NULL;   /* Index pointer to step along rotation list array */
  char *reserved_ads_list_base     = NULL;   /* Char strings and integers to parse ad_reservations file */
  char *reserved_ads_list_base_ptr = NULL;
  int   ad_info_offset             = 0;
  char *tmp_pagename               = NULL;
  char *tmp_ad_position            = NULL;
  int   ad_number                  = 0;	     /* Number of ads read from current list */
  int   ad_position                = 0;      /* Ad position number based on MAX_AD_POSITIONS */
  char *filepath                   = NULL;   /* string of the full path to the file being read */
  int   pid                        = 0;      /* Current child process ID for logging child_init error messages */


  pid = getpid();


  /* For each VirtualHost load ad_info table, ad rotations, and reserved pages list */

  for( server=main_server; server; server = server->next ) {

#if ABDEBUG_VERBOSE

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[%d]: Processing AdBanner files for %s", 
		   pid,
		   server->server_hostname);
#endif


    sconf = (adbanners_server_rec *) ap_get_module_config(server->module_config, &adbanners_module);


  /* Get the full path of the AdBannerInfoFile so we can read the file contents */

    if ((filepath = sconf->ad_info_filepath) == NULL) {

#if ABDEBUG

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[%d]: Skipping vhost %s ad_info_filepath is NULL",
		   pid,
		   server->server_hostname);
#endif

      continue;  /* next vhost */

    } /* if ad_info_filepath */



    /* Check to make sure the given path argument is a valid path to a file of reasonable size... */
    /* If that's not a valid file then log the error and leave the ad_1_rotate list pointer NULL so we ignore it */

    if (stat(filepath, &finfo) == -1) {

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_ERR, 
		   server,
		   "%s: adbanners_child_init[%d] Can't read AdBannerInfoFile \"%s\": %s",
		   __FILE__,
		   pid,
		   filepath, 
		   strerror(errno));
      continue;  /* next vhost */

    }  /* if stat filepath */


    if (finfo.st_size > MAX_ADINFO_FILESIZE) {

#if ABDEBUG

	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     server,
		     "%s: adbanners_child_init[%d] AdBannerInfoFile \"%s\" is too big (%ld bytes) so I'm only going to read the first %d bytes from it.", 
		     __FILE__,
		     pid,
		     filepath, 
		     finfo.st_size, 
		     MAX_ADINFO_FILESIZE);

#endif

	adinfo_size = MAX_ADINFO_FILESIZE;


    } else if (finfo.st_size < ADID_SIZE) {


#if ABDEBUG

	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_ERR, 
		     server,
		     "%s: adbanners_child_init[%d] AdBannerInfoFile \"%s\" is too small (%ld bytes) so %s is ignoring it.", 
		     __FILE__,
		     pid,
		     filepath, 
		     finfo.st_size, 
		     __FILE__);

#endif

	continue;  /* next vhost */
	
    } /* if finfo.st_size */


    adinfo_size = finfo.st_size;

    
    /* 
     Allocate space for contents of ad_info.dat file in the server heap.
     We allocate (ad_info_size + 1) and use calloc() to assure that our ad_info table string will be NULL terminated.
    */

    if ((adinfo_base = ap_pcalloc(p, (adinfo_size + 1) * sizeof(char))) == NULL) {
      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "%s: adbanners_child_init[%d] Apache would not allocate %d bytes of memory to hold ad_info table.", 
		   __FILE__,
		   pid,
		   adinfo_size);
      continue;  /* next vhost */
    }


#if ABDEBUG_VERBOSE

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "%s: adbanners_child_init[%d] Allocated %d bytes of memory to hold ad_info table.", 
		   __FILE__,
		   pid,
		   adinfo_size);

#endif



#if SLURP_ADINFO

    /* Slurp the entire ad_info file into allocated memory block */

    if ((indexer = ap_popenf(p, filepath, O_RDONLY, 0)) < 1) {

      read(fd, adinfo_base, adinfo_size);
      ap_pclosef(p, indexer);

    } else {
      
      /* TODO: log an error message here */

    }

#endif 



    /* Allocate a temporary buffer space for ad_info_line */
  
    ad_info_line = ap_pcalloc(p, (MAX_LINE_LENGTH + 2) * sizeof(char));

    /* Grok contents of ad_info.dat file into the allocated memory block */

    adinfo = adinfo_base;
    indexer = 0;

    fp = ap_pfopen(p, filepath, "r");
    while (fgets(line_buffer, MAX_LINE_LENGTH, fp) != NULL) {
	
      sprintf(ad_info_line, "%s\n", line_buffer);
      indexer += strlen(ad_info_line) - 1;
      sprintf(adinfo, ad_info_line);
      adinfo = adinfo_base + (sizeof(char) * indexer);


#if ABDEBUG_VERBOSE

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[%d] adinfo ptr = (adinfo_base (%d) + indexer (%d)) = %d", 
		   pid, 
		   adinfo_base, 
		   indexer, 
		   adinfo);
#endif



      /* Avoid reading more than our reasonable limit for ad_info table size */
      if (indexer >= adinfo_size) {
	break;
      }

    } /* while line_buffer */

    ap_pfclose(p, fp);



    /* Insert pointer to the ad_info table string in the server config record */
    sconf->ad_info_base = adinfo_base;


#if ABDEBUG_VERBOSE

    ap_log_error(APLOG_MARK, 
		 APLOG_NOERRNO | APLOG_NOTICE, 
		 server,
		 "adbanners_child_init[%d] Finished reading ad_info table file: %ld Table base address = %ld, sconf->ad_info_base = %ld Table stop address = %ld, adinfo_size = %d, strlen(adinfo_base) = %d bytes, indexer = %d",
		 pid,
		 filepath, 
		 adinfo_base, 
		 sconf->ad_info_base, 
		 adinfo,
		 adinfo_size,
		 strlen(adinfo_base),
		 indexer);

#endif


    /* Make sure we have a valid ad_info table pointer in memory before attempting to dereference it */

    if (sconf->ad_info_base == NULL) {

#if ABDEBUG

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_ERR, 
		   server,
		   "%s adbanners_child_init[%d] Not processing AdRotate files: ad_info_base is NULL: AdBannerInfoFile was not loaded.",
		   __FILE__, 
		   pid);

#endif

      continue;  /* next vhost */

    }


    /* for each ad position, read the AdBannerRotateFile contents... */

    for (ad_position = 1; ad_position < MAX_AD_POSITIONS; ad_position++) {

      if ((filepath = sconf->ad_x_rotate_filepath[ad_position]) == NULL) {

#if ABDEBUG_VERBOSE

	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     server,
		     "adbanners_child_init[%d]: ad_x_rotate_filepath[%d] is NULL.", 
		     pid, 
		     ad_position);
#endif

	continue;   /* next ad_position */

      }

    
      /* Check to make sure the given path argument is a valid path to a file of reasonable size... */
      /* If that's not a valid file then log the error and leave the ad_1_rotate list pointer NULL so we ignore it */

      if (stat(filepath, &finfo) == -1) {

	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_ERR, 
		     server,
		     "%s adbanners_child_init[pid%d]: Can't read file \"%s\" given in AdBannerRotateFile %d directive: %s",
		     __FILE__,
		     pid,
		     filepath,
		     ad_position,
		     strerror(errno));
	continue;   /* next ad position */

      }  /* if stat filepath */


      if (finfo.st_size > MAX_ROTATION_FILESIZE) {

#if ABDEBUG

	  ap_log_error(APLOG_MARK, 
		       APLOG_NOERRNO | APLOG_NOTICE, 
		       server,
		       "%s adbanners_child_init[%d]: AdBannerRotateFile \"%s\" too big (%ld bytes) so only reading the first %d bytes from it.",
		       __FILE__, 
		       pid,
		       filepath, 
		       finfo.st_size, 
		       MAX_ROTATION_FILESIZE);

#endif


	  listsize = MAX_ROTATION_LISTSIZE;	/* set to 1024 items */
	  

      } else if (finfo.st_size < ADID_SIZE) {


#if ABDEBUG_VERBOSE

	  ap_log_error(APLOG_MARK, 
		       APLOG_NOERRNO | APLOG_NOTICE, 
		       server,
		       "%s adbanners_child_init[%d]: AdBannerRotateFile \"%s\" is too small (%ld bytes) so %s is ignoring it.", 
		       __FILE__, 
		       pid,
		       filepath, 
		       finfo.st_size, 
		       __FILE__);

#endif

	  continue;  /* next ad_position */

      }  /* if finfo.st_size */



      /* Estimate of number of list elements based on rotation file size */
      listsize = (int) (finfo.st_size / (ADID_SIZE + 1));



      /* Allocate a block of contigous space to store the processed rotation list */
      /* The allocation space is being based on number of lines in rotation list file so at worst it will be overallocated */

      rotation_list_base = NULL;

      if ((rotation_list_base = ap_pcalloc(p, ((listsize + 1) * sizeof(int)))) == NULL) {

	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_ERR, 
		     server,
		     "%s: Apache would not allocate %d bytes of memory to hold position %d rotation list.", 
		     __FILE__, 
		     ((listsize + 1) * sizeof(int)), 
		     ad_position);
	continue;  /* next ad_position */

      }

      rotation_list_base_ptr = rotation_list_base;


#if ABDEBUG_VERBOSE

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[%d] Reading position [%d] rotation file: %s...", pid, ad_position, filepath);
      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[%d] Allocated %d bytes of server memory pool to store rotation list of %d integers", 
		   pid,
		   ((listsize + 1) * sizeof(int)), 
		   listsize);
      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[%d] Reading address of ad_info table sconf->ad_info_base = %d", 
		   pid, 
		   sconf->ad_info_base);
      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[%d] Matching rotation list [%d] ads to ad_info table rows:", 
		   pid, 
		   ad_position);
      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[%d] ad_number\t\tad_info Offset\tad_id\n----------------------------------------------------", 
		   pid);

#endif


      /* Read each line in the ad_rotate list */

      ad_number = 0;


      fp = ap_pfopen(p, filepath, "r");

      while (fgets(line_buffer, MAX_LINE_LENGTH, fp) != NULL) {

	ad_info = sconf->ad_info_base;

	/* Using strtok() to get rid of nasty trailing carriage return */

	ad_id = strtok(line_buffer, "\n\r");


	/* Skip blank lines */

	if ( ad_id == NULL ) {
#if ABDEBUG_VERBOSE
	  ap_log_error(APLOG_MARK, 
		       APLOG_NOERRNO | APLOG_NOTICE, 
		       server,
		       "adbanners_child_init[%d] SKIP: ad_position %d, ad_number %d: ad_id NULL", 
		       pid,
		       ad_position,
		       ad_number);
#endif

	  continue;  /* next line_buffer */
	}


#if ABDEBUG_VERBOSE

	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     server,
		     "adbanners_child_init[%d] MARK: ad_position %d, about to strlen [%s]...", 
		     pid,
		     ad_position,
		     ad_id);
#endif


	if (strlen(ad_id) != ADID_SIZE) {

#if ABDEBUG_VERBOSE
	  ap_log_error(APLOG_MARK, 
		       APLOG_NOERRNO | APLOG_NOTICE, 
		       server,
		       "adbanners_child_init[%d] SKIP: ad_position %d, ad_number %d: ad_id string wrong size", 
		       pid,
		       ad_position,
		       ad_number);
#endif

	  continue;  /* next line_buffer */
	}


	/* Using strstr() to search for this ad_id in our ad_info table string */

	ad_info_ptr = strstr(ad_info, ad_id);

	if (ad_info_ptr != NULL) {

#if ABDEBUG_VERBOSE
	  ap_log_error(APLOG_MARK, 
		       APLOG_NOERRNO | APLOG_NOTICE, 
		       server,
		       "adbanners_child_init[%d] MARK: ad_position %d, about to set rotation_list_base_ptr...", 
		       pid,
		       ad_position);
#endif


	  /* Push the offset of this row in ad_info table onto our rotation list */

	  *rotation_list_base_ptr = (ad_info_ptr - ad_info);

#if ABDEBUG_VERBOSE
	  ap_log_error(APLOG_MARK, 
		       APLOG_NOERRNO | APLOG_NOTICE, 
		       server,
		       "adbanners_child_init[%d] Position [%d] Ad [%d]\t=\t%d (%s...)",
		       pid,
		       ad_position, 
		       ad_number, 
		       *rotation_list_base_ptr, 
		       ad_id);
#endif

	  ad_number++;
	  rotation_list_base_ptr++;

	  /* Stop reading ad_x_rotate.dat if we've read too many lines already */

	  if (ad_number >= MAX_ROTATION_LISTSIZE) {
	    break;
	  }

	}  /* if ad_info_ptr */


      }	/* while fgets() ad_x_rotate.dat */


      ap_pfclose(p, fp);



      /* Insert pointer to ad_x_rotate list and ad_x_numberofads in the server config record */
      sconf->ad_x_rotate[ad_position]      = rotation_list_base;
      sconf->ad_x_numberofads[ad_position] = ad_number;


#if ABDEBUG_VERBOSE
      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[%d] Position [%d], numberofads = %d, rotation_list_base address = %d Position [%d] rotation_list_base_ptr (stop address) = %d",
		   pid,
		   ad_position,
		   ad_number,
		   rotation_list_base, 
		   ad_position, 
		   rotation_list_base_ptr);
#endif



    } /* for each ad_position, load rotation list */



    /* Now read the ad_reservations table that reserves certain ads to certain web pages in certain positions */

    listsize = 0;

    if ((filepath = sconf->ad_reserved_pages_filepath) == NULL) {

#if ABDEBUG

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[pid%d]: Not processing reserved pages file. sconf->ad_reserved_pages_filepath is NULL\n", 
		   pid);
#endif

    } else {

      if (stat(filepath, &finfo) == -1) {

#if ABDEBUG

	ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_ERR, 
		   server,
		   "%s: adbanners_child_init[%d] Can't read AdBannerReservedPagesFile \"%s\": %s",
		   __FILE__,
		   pid,
		   filepath, 
		   strerror(errno));

#endif


      } else {

	if (finfo.st_size > MAX_ROTATION_FILESIZE) {

#if ABDEBUG

	  ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     server,
		     "%s: adbanners_child_init[%d] AdBannerReservedPagesFile \"%s\" is too big (%ld bytes) so only reading the first %d bytes from it.", 
		     __FILE__,
		     pid,
		     filepath, 
		     finfo.st_size, 
		     MAX_ROTATION_FILESIZE);
      
#endif

	  listsize = MAX_ROTATION_FILESIZE;	/* set to 1024 items */
      
	} else {

	  /* Estimate of number of elements in list based on rotation file size */
	  listsize = finfo.st_size;

	} /* if st_size */
    
      }  /* if stat reserved pages filepath */

    } /* if reserved_pages_filepath */


    /* Allocate a char string big enough to hold processed reserved pages list */

    if (listsize < ADID_SIZE) {

#if ABDEBUG_VERBOSE

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "%s: adbanners_child_init[%d] AdBannerReservedPagesFile \"%s\" is too small (%ld bytes) to bother processing.",
		   __FILE__,
		   pid,
		   filepath, 
		   finfo.st_size);
#endif


    } else {

      if ((reserved_ads_list_base = ap_pcalloc(p, (listsize * sizeof(char)))) == NULL) {

#if ABDEBUG
	
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_ERR, 
		     server,
		     "%s: adbanners_child_init[%d] Apache would not allocate %d bytes of memory to hold reserved pages list.", 
		     __FILE__,
		     pid,
		     (listsize * sizeof(char)));
#endif


      } else {

	reserved_ads_list_base_ptr = reserved_ads_list_base;


	/* Read each line in the ad_reservations list file */

	ad_number = 0;
  
	fp = ap_pfopen(p, filepath, "r");
	while (fgets(line_buffer, MAX_LINE_LENGTH, fp) != NULL) {

	  ad_info = sconf->ad_info_base;

	  /* Don't bother reading empty lines */

	  if (strlen(line_buffer) < ADID_SIZE) {

#if ABDEBUG_VERBOSE

	    ap_log_error(APLOG_MARK, 
			 APLOG_NOERRNO | APLOG_NOTICE, 
			 server,
			 "adbanners_child_init[%d] skipping blank line in ad reserved pages file.", 
			 pid);
#endif
	    continue;  /* next line */
	  }


	  /* Using strtok() to break line buffer into tokens */

	  tmp_pagename    = strtok(line_buffer, "\t");
	  tmp_ad_position = strtok(NULL, "\t");
	  ad_id           = strtok(NULL, "\n");

#if ABDEBUG_VERBOSE

	  ap_log_error(APLOG_MARK, 
		       APLOG_NOERRNO | APLOG_NOTICE, 
		       server,
		       "adbanners_child_init[%d] Reserved page file line %d: %s, %s, %s", 
		       pid,
		       ad_number, 
		       tmp_pagename, 
		       tmp_ad_position, 
		       ad_id);
#endif


      /* Using strstr() to search for this ad_id in our ad_info table string */

	  ad_info_ptr = strstr(ad_info, ad_id);

	  if (ad_info_ptr != NULL) {

	    ad_info_offset = (ad_info_ptr - ad_info);
	    ad_position    = atoi(tmp_ad_position);

	    if ((ad_position < 1) || (ad_position > MAX_AD_POSITIONS)) {

#if ABDEBUG

	      ap_log_error(APLOG_MARK, 
			   APLOG_NOERRNO | APLOG_NOTICE, 
			   server,
			   "adbanners_child_init[%d] Reserved ad position (%d) is out of range. Ignoring.", 
			   pid,
			   ad_position);
#endif
	      /* TODO: Test with bogus ad_reservations data to make sure this jumps back to start of while loop */
	      continue; /* next line */
	    }


	    sprintf(line_buffer, "%s\t%d\t%d\n", tmp_pagename, ad_position, ad_info_offset);
	    sprintf(reserved_ads_list_base_ptr, "%s", line_buffer);


#if ABDEBUG_VERBOSE

	    ap_log_error(APLOG_MARK, 
			 APLOG_NOERRNO | APLOG_NOTICE, 
			 server,
			 "adbanners_child_init[%d] Reserved page [%d] %s", 
			 pid,
			 ad_number, 
			 line_buffer);
#endif

	    ad_number++;
	    reserved_ads_list_base_ptr += (strlen(line_buffer) * sizeof(char));


	    /* Stop reading ad_reservations.dat if we've read too many lines already */
      
	    if (strlen(reserved_ads_list_base) >= MAX_ROTATION_FILESIZE) {

#if ABDEBUG
	      ap_log_error(APLOG_MARK, 
			   APLOG_NOERRNO | APLOG_NOTICE, 
			   server,
			   "adbanners_child_init[%d] Stopped reading ad reservations file at (%d byte) limit.",
			   pid,
			   MAX_ROTATION_FILESIZE);
#endif
	      break;
	    }

	  } else {

#if ABDEBUG
	    ap_log_error(APLOG_MARK, 
			 APLOG_NOERRNO | APLOG_NOTICE, 
			 server,
			 "adbanners_child_init[%d] Reserved page %s, ad_id (%s) is not in ad_info table. Ignoring.", 
			 pid,
			 tmp_pagename,
			 ad_id);
#endif

	  } /* if ad_info_ptr */

	}  /* while fgets() ad_reservations.dat */
  
	ap_pfclose(p, fp);



#if ABDEBUG_VERBOSE

	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     server,
		     "adbanners_child_init[%d] reservedpages_list_base address = %d reserved_ads_list_base_ptr (stop address) = %d",
		     pid,
		     reserved_ads_list_base, 
		     reserved_ads_list_base_ptr);
	
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     server,
		     "adbanners_child_init[%d] Processed reserved pages list (%d bytes): %d", 
		     pid,
		     strlen(reserved_ads_list_base), 
		     reserved_ads_list_base);
	
#endif
  

      } /* if calloc reserved_ads_list_base */

    } /* if reserved pages listsize */




    /* Save reserved pages list to our server config record */

    sconf->ad_reserved_pages = reserved_ads_list_base;




#if GEOTARGETING

    listsize = 0;

    /* Now read the ad_reservations table that reserves certain ads to certain web pages in certain positions */

    if ((filepath = sconf->ad_reserved_domains_filepath) == NULL) {

#if ABDEBUG

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[pid%d]: Not processing reserved domains file. sconf->ad_reserved_pages_filepath is NULL\n", 
		   pid);
#endif

      continue;

    }


    if (stat(filepath, &finfo) == -1) {

#if ABDEBUG

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_ERR, 
		   server,
		   "%s: adbanners_child_init[%d] Can't read AdBannerReservedDomainsFile \"%s\": %s",
		   __FILE__,
		   pid,
		   filepath, 
		   strerror(errno));

#endif

      continue;

    } else {

      if (finfo.st_size > MAX_ROTATION_FILESIZE) {

#if ABDEBUG

	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     server,
		     "%s: adbanners_child_init[%d] AdBannerReservedDomainsFile \"%s\" is too big (%ld bytes) so only reading the first %d bytes from it.", 
		     __FILE__,
		     pid,
		     filepath, 
		     finfo.st_size, 
		     MAX_ROTATION_FILESIZE);
      
#endif

	listsize = MAX_ROTATION_FILESIZE;	/* set to 1024 items */
      
      } else {

	/* Estimate of number of elements in list based on rotation file size */

	listsize = finfo.st_size;

      }
    
    }  /* if stat reserved pages filepath */


    /* Allocate a char string big enough to hold processed reserved pages list */

    if (listsize > ADID_SIZE) {

      if ((reserved_ads_list_base = ap_pcalloc(p, (listsize * sizeof(char)))) == NULL) {

#if ABDEBUG
	
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_ERR, 
		     server,
		     "%s: adbanners_child_init[%d] Apache would not allocate %d bytes of memory to hold reserved domains list.", 
		     __FILE__,
		     pid,
		     (listsize * sizeof(char)));
#endif

	continue;
      }

      reserved_ads_list_base_ptr = reserved_ads_list_base;

    } else {

#if ABDEBUG_VERBOSE

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "%s: adbanners_child_init[%d] AdBannerReservedDomainsFile \"%s\" is too small (%ld bytes) to bother processing.",
		   __FILE__,
		   pid,
		   filepath, 
		   finfo.st_size);
#endif
      continue;

    } /* if reserved domains listsize */


    /* Make sure we have a valid ad_info table pointer in memory before attempting to dereference it */

    if (sconf->ad_info_base == NULL) {

#if ABDEBUG

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_ERR, 
		   server,
		   "%s: adbanners_child_init[%d] Can't process AdBannerReservedDomainsFile. AdBannerInfo table is NULL",
		   __FILE__,
		   pid);

#endif

      continue;
    }


    /* Read each line in the ad_reservations list file */

    ad_number = 0;
  
    fp = ap_pfopen(p, filepath, "r");
    while (fgets(line_buffer, MAX_LINE_LENGTH, fp) != NULL) {

      ad_info = sconf->ad_info_base;


      /* Don't bother reading empty lines */

      if (strlen(line_buffer) < ADID_SIZE) {

#if ABDEBUG_VERBOSE

	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     server,
		     "adbanners_child_init[%d] skipping blank line in ad reserved domains file.", 
		     pid);
#endif
	continue;
      }


      /* Using strtok() to break line buffer into tokens */

      tmp_pagename    = strtok(line_buffer, "\t");
      tmp_ad_position = strtok(NULL, "\t");
      ad_id           = strtok(NULL, "\n");

#if ABDEBUG_VERBOSE

      ap_log_error(APLOG_MARK, 
		   APLOG_NOERRNO | APLOG_NOTICE, 
		   server,
		   "adbanners_child_init[%d] Reserved domain file line %d: %s, %s, %s", 
		   pid,
		   ad_number, 
		   tmp_pagename, 
		   tmp_ad_position, 
		   ad_id);
#endif


      /* Using strstr() to search for this ad_id in our ad_info table string */

      ad_info_ptr = strstr(ad_info, ad_id);

      if (ad_info_ptr != NULL) {

	ad_info_offset = (ad_info_ptr - ad_info);

	ad_position = atoi(tmp_ad_position);
      
	if ((ad_position < 1) || (ad_position > MAX_AD_POSITIONS)) {

#if ABDEBUG

	  ap_log_error(APLOG_MARK, 
		       APLOG_NOERRNO | APLOG_NOTICE, 
		       server,
		       "adbanners_child_init[%d] Reserved ad position (%d) is out of range. Ignoring.", 
		       pid,
		       ad_position);
#endif
	  /* TODO: Test with bogus ad_reservations data to make sure this jumps back to start of while loop */
	  continue;
	}


	sprintf(line_buffer, "%s\t%d\t%d\n", tmp_pagename, ad_position, ad_info_offset);
	sprintf(reserved_ads_list_base_ptr, "%s", line_buffer);


#if ABDEBUG_VERBOSE

	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     server,
		     "adbanners_child_init[%d] Reserved domain [%d] %s", 
		     pid,
		     ad_number, 
		     line_buffer);
#endif

	ad_number++;
	reserved_ads_list_base_ptr += (strlen(line_buffer) * sizeof(char));


	/* Stop reading ad_reservations.dat if we've read too many lines already */
      
	if (strlen(reserved_ads_list_base) >= MAX_ROTATION_FILESIZE) {

#if ABDEBUG
	  ap_log_error(APLOG_MARK, 
		       APLOG_NOERRNO | APLOG_NOTICE, 
		       server,
		       "adbanners_child_init[%d] Stopped reading ad reserved domains file at (%d byte) limit.",
		       pid,
		       MAX_ROTATION_FILESIZE);
#endif
	  break;
	}

      } else {

#if ABDEBUG
	ap_log_error(APLOG_MARK, 
		     APLOG_NOERRNO | APLOG_NOTICE, 
		     server,
		     "adbanners_child_init[%d] Reserved domain %s, ad_id (%s) is not in ad_info table. Ignoring.", 
		     pid,
		     tmp_pagename,
		     ad_id);
#endif

      }

    }  /* while fgets() ad_reserved_domains.dat */
  
    ap_pfclose(p, fp);



#if ABDEBUG_VERBOSE

    ap_log_error(APLOG_MARK, 
		 APLOG_NOERRNO | APLOG_NOTICE, 
		 server,
		 "adbanners_child_init[%d] reserveddomains_list_base address = %d reserved_ads_list_base_ptr (stop address) = %d",
		 pid,
		 reserved_ads_list_base, 
		 reserved_ads_list_base_ptr);

    ap_log_error(APLOG_MARK, 
		 APLOG_NOERRNO | APLOG_NOTICE, 
		 server,
		 "adbanners_child_init[%d] Processed reserved domains list (%d bytes): %d", 
		 pid,
		 strlen(reserved_ads_list_base), 
		 reserved_ads_list_base);

#endif
  

    /* Save reserved domains list to our server config record */

    sconf->ad_reserved_domains = reserved_ads_list_base;

#endif
    /* end if GEOTARGETING */


  } /* foreach vhost  */


} /* end of adbanners_child_init() */







/*
   This function is called with every HTTP request. A few things happen here:

   1. Request is immediately returned back to the caller if this request was not an HTML page
   2. The reserved pages table is checked to see if the URI for this request is a reserved page
   3. If a reserved ad was not found for this page then the next ad in each ad position is selected randomly
   4. The position 2 ad is preset to megabanner part B if position 1 is a megabanner
   5. The HTML for each ad in the page is stored in a separate environment variable
   6. The request is handed back to the caller for further processing by other Apache modules

 */



static int adbanners_rotate(request_rec *r)
{

    adbanners_server_rec *sconf         =  NULL;     /* Pointer to the server config record (or seemingly a child process copy of it... grrrr...) */
    char *ad_info_ptr1                  =  NULL;
    char *ad_info_ptr2                  =  NULL;     /* Copy of sconf->ad_info_base */
    int *rotation_list_base             =  NULL;     /* Copy of sconf->ad_x_rotate list base address */
    char *reservedpages_list_base       =  NULL;     /* Copy of sconf->ad_reserved_pages list base address */
    char *ad_reserved_list_ptr1         =  NULL;     /* A pointer to a specific row in reserved pages list */
    char *tmp_reserved_line             =  NULL;     /* A temporary string to hold the pattern we'll try to find in reserved pages list */
    char tmp_reserved_adid_offset_buffer[MAX_ADID_OFFSET_SIZE];	/* A small char buffer to translate reserved adid from ascii to int */
    int ad_number                       =  0;
    int ad_offset                       =  0;	     /* The pointer offset from base of ad_info table for current ad_id */
    int ad_numberofads                  =  0;	     /* Rotation list index and rotation list number of elements */
    int ad_info_line_length             =  0;	     /* strlen() of ad_info_line */
    int i                               =  0;	     /* An all-purpose loop counter */
    int tmp_time                        =  0;        /* Return value of time(NULL) server's epoch time */
    int ad_position                     =  1;	     /* A loop counter to step through each ad position (1..3) */
    char tmp_ad_info_line_buffer[MAX_LINE_LENGTH];   /* Temporary string to hold ad_info data string for each ad */
    char *tmp_ad_banner_adid            =  NULL;     /* Temporary string to read AdID */
    char *tmp_ad_banner_width           =  NULL;
    char *tmp_ad_banner_height          =  NULL;
    char *tmp_ad_banner_target_url      =  NULL;
    char *tmp_ad_banner_alt_text        =  NULL;
    char *tmp_ad_status                 =  NULL;     /* Temporary buffer holding current module status string */
    char *tmp_ad_text                   =  NULL;     /* Temporary buffer holding the raw HTML code of each ad banner */
    char *tmp_ad_env_label              =  NULL;     /* A text string which labels each ENV variable we set */
    char *tmp_cachebuster               =  NULL;     /* The current server epoch time in the form of a char string */
    char tmp_ad_type                    =  '0';	     /* A single digit which will be '5' or '6' if the ad banner is an HTML layer */
    char *tmp_htmlad_id[ADID_SIZE + 1];		     /* A buffer to the ad_id of an HTML so we can derive its dat filename */
    char *tmp_htmlad_filepath           =  NULL;     /* A temporary buffer in which to build the full path to HTML file for 5000 ads */
    char tmp_megabanner_id[ADID_SIZE + 1];	     /* A buffer to hold the ad_id for when a megabanner is served */
    int fd                              =  0;	     /* A file pointer for reading contents HTML ad layer files */
    char *tmp_htmlad_filecontents       =  NULL;     /* A char string containing file contents */
    char *tmp_htmlad_filecontents_ptr1  =  NULL;     /* A pointer to step through tmp_htmlad_filecontents to replace INSERT_CACHEBUSTER */
    char *tmp_extra_text                =  NULL;     /* A char string for pagecounter HTML code or other HTML code per ad position */
    char *tmp_htmlad_imageurl           =  NULL;     /* AdBannerImageUrl + ad_id used to determine if htmlad already has a banner image tag */
    char tmp_pagename_buffer[MAX_PAGENAME_LENGTH];   /* Filename of the requested web page */

#if IFRAME_SERVER
    char *tmp_uri                       =  NULL;
    uri_components referer_uri;                      /* A uri structure for dissecting HTTP_REFERER URL into hostname and file path */
    const char *tmp_referer;
#endif

    char *ad_user_remote_host           =  NULL;     /* User's remote hostname from reverse IP lookup */
    char *ad_user_domain                =  NULL;     /* User's two-char domain code from last two char's of remote hostname */
#if GEOTARGETING
    char *reserveddomains_list_base     =  NULL;     /* Copy of sconf->ad_reserved_domains list base address */
    char ad_user_geocode0               =  '1';      /* The last few chars in user's remote host name */
    char ad_user_geocode1               =  '1';
    char ad_user_geocode2               =  '1';
#endif


#if PROCESS_ALL_REQUESTS
#else


    /* If this request is not a web page then sit back, relax, and feel your ass grow. */

    if ( !( strstr(r->uri, ".html" ) || strstr( r->uri, ".pl" ) || strstr( r->uri, ".shtml" ) || strstr( r->uri, ".php") ) ) {
	return DECLINED;
    }

#endif



    /* Don't bother with HEAD requests */

    if (r->header_only) {
	return DECLINED;
    }



    /* Get a pointer to our module data record */

    sconf = (adbanners_server_rec *) ap_get_module_config(r->server->module_config, &adbanners_module);



    /* Generate epoch timestamp to be used as server-side cachebuster for impression counter */

    tmp_time = time(NULL);
    tmp_cachebuster = ap_psprintf(r->pool, "%d", tmp_time);



    /* Get the base filename of this web page and server epoch time to insert in ad HTML code */

    memset( tmp_pagename_buffer, 0, MAX_PAGENAME_LENGTH );
    if ( r->uri ) {
      i = ap_ind( r->uri, '.' ) - 1;
      if ( i < 0 ) {
	i = strlen( r->uri );
      }
      memcpy( tmp_pagename_buffer, (char *) ( r->uri + 1 ), i );
      i = strlen( tmp_pagename_buffer );
      if ( i < 1 ) {         /* if pagename == '/' then assume it's 'index.html' */
	memcpy( tmp_pagename_buffer, "index", strlen( "index" ) );


#if IFRAME_SERVER

      } else if ( i == 1 ) { /* if pagename == 'x.html' then  assume we're on an IFRAME server, get pagename from HTTP_REFERER */

	tmp_referer = ap_table_get( r->headers_in, "Referer" );

	if ( tmp_referer ) {
	  if ( ( ap_parse_uri_components( r->pool, tmp_referer, &referer_uri ) ) == HTTP_OK ) {
	    tmp_uri = referer_uri.path;
	    if ( tmp_uri ) {
	      i = ap_ind( tmp_uri, '.' ) - 1;
	      if ( i < 0 ) {
		i = strlen( tmp_uri );
	      }
	      memcpy( tmp_pagename_buffer, (char *) ( tmp_uri + 1 ), i );
	      i = strlen( tmp_pagename_buffer );
	      if ( i < 1 ) {         /* if pagename == '/' then assume it's 'index.html' */
		memcpy( tmp_pagename_buffer, "index", strlen( "index" ) );
	      }
	    } /* if referer filepath */
	  } /* if parse_uri */
	} /* if referer */
#endif

      } /* if i < 1 */
    } /* if pagename */
    

    /* Generate the pagecount tag in ad position 1 */

    if ( sconf->ad_pagecount_url ) {

      if ( sconf->ad_pagecount_url2 ) {

	tmp_extra_text = ap_psprintf(r->pool,

"<TABLE BORDER=0 CELLPADDING=0 CELLSPACING=0><TR><TD WIDTH=2><SCRIPT LANGUAGE=\"JAVASCRIPT\">\n\
<!--\n\
now = new Date();\n\
tail = now.getTime();\n\
document.write(\"<IMG SRC=\'%s\?%s,\");\n\
document.write(tail);\n\
document.write(\"\' WIDTH=1 HEIGHT=1 BORDER=0>\");\n\
document.write(\"<IMG SRC=\'%s\?%s,\");\n\
document.write(tail);\n\
document.write(\"\' WIDTH=1 HEIGHT=1 BORDER=0><BR>\");\n\
//-->\n\
</SCRIPT>\n\
<NOSCRIPT>\n\
<IMG SRC=\"%s\?%s,%s\" WIDTH=1 HEIGHT=1 BORDER=0>\n\
<IMG SRC=\"%s\?%s,%s\" WIDTH=1 HEIGHT=1 BORDER=0><BR>\n\
</NOSCRIPT></TD></TR></TABLE>\n", 
				     sconf->ad_pagecount_url, 
				     tmp_pagename_buffer,
				     sconf->ad_pagecount_url2, 
				     tmp_pagename_buffer, 
				     sconf->ad_pagecount_url, 
				     tmp_pagename_buffer, 
				     tmp_cachebuster,
				     sconf->ad_pagecount_url2, 
				     tmp_pagename_buffer, 
				     tmp_cachebuster );

      } else {

	tmp_extra_text = ap_psprintf(r->pool,

"<TABLE BORDER=0 CELLPADDING=0 CELLSPACING=0><TR><TD WIDTH=1><SCRIPT LANGUAGE=\"JAVASCRIPT\">\n\
<!--\n\
now = new Date();\n\
tail = now.getTime();\n\
document.write(\"<IMG SRC=\'%s\?%s,\");\n\
document.write(tail);\n\
document.write(\"\' WIDTH=1 HEIGHT=1 BORDER=0><BR>\");\n\
//-->\n\
</SCRIPT>\n\
<NOSCRIPT>\n\
<IMG SRC=\"%s\?%s,%s\" WIDTH=1 HEIGHT=1 BORDER=0><BR>\n\
</NOSCRIPT></TD></TR></TABLE>\n",
				     sconf->ad_pagecount_url, 
				     tmp_pagename_buffer, 
				     sconf->ad_pagecount_url, 
				     tmp_pagename_buffer, 
				     tmp_cachebuster );


      }  /* if defined ad_pagecount_url2  */

      ap_table_setn(r->subprocess_env, "AD_BANNER_1", tmp_extra_text);

    }  /* if defined pagecount_url */


#if MEGABANNERS

    /* Just in case the position 2 rotation runs empty we'll seed the navbar text here anyway */

    if ( sconf->ad_navbar_html ) {

	      tmp_ad_text = ap_psprintf(r->pool, 
"\n %s \n<IMG SRC=\"%s/top.gif\" WIDTH=135 HEIGHT=10><BR>\n\
<TABLE CELLPADDING=0 CELLSPACING=0 WIDTH=135 BORDER=0>\
<TR><TD BGCOLOR=\"#%s\" WIDTH=135 ALIGN=CENTER VALIGN=TOP>\
<FONT FACE=\"ARIAL,HELVETICA\" SIZE=2> <!-- HTML for ad 2 was not generated. See ADBANNER_STATUS output for details. --> </FONT>\n\
</TD></TR></TABLE><IMG SRC=\"%s/bottom.gif\" WIDTH=135 HEIGHT=10><BR><BR>\n\n", 
					sconf->ad_navbar_html, 
					sconf->ad_navbar_gifsurl, 
					sconf->ad_navbar_bgcolor, 
					sconf->ad_navbar_gifsurl);

	      ap_table_setn(r->subprocess_env, "AD_BANNER_2", tmp_ad_text);

    }

#endif


    /* Make sure we have an ad_info table pointer before proceeding */

    if ( sconf->ad_info_base == NULL ) {
	tmp_ad_status = ap_psprintf(r->pool, "<B>%s:</B> No ad_info table loaded. To find out why check error_log for AdBanner errors.", __FILE__);
	ap_table_setn(r->subprocess_env, "ADBANNER_STATUS", tmp_ad_status);
	return DECLINED;
    }



#if ABDEBUG_VERBOSE
    tmp_ad_status = ap_psprintf(r->pool, "<B>%s:</B> ad_info dump:\n%s\n", __FILE__, sconf->ad_info_base);
    ap_table_setn(r->subprocess_env, "ADBANNER_STATUS", tmp_ad_status);
#endif



    /* Clear megabanner id buffer (will be set if position 1 turns out to be a megabanner) */

    memset(tmp_megabanner_id, 0, ADID_SIZE + 1);



#if GEOTARGETING

    /* If we have some geotargeted ads then read user's remote domain name */

    reserveddomains_list_base = sconf->ad_reserved_domains;

    if ( reserveddomains_list_base != NULL ) {

      /* Calling ap_get_remote_host() to get remote user's hostname (domain)
	 The third argument of ap_get_remote_host() can be 1, 2, or 3 depending 
	 on if HostnameLookups is On or Off. If HostnameLookups is Off then you have 
	 to pass 3 as the third argument which forces a double-reverse lookup and 
	 that's not good, you're better off just having HostnameLookups On in which
	 case setting up a local caching DNS server on the same machine helps.
      */

      ad_user_remote_host = ap_psprintf(r->pool, "%s", ap_get_remote_host(r->connection, r->per_dir_config, 1));

      i = strlen ( ad_user_remote_host );    
      if ( i > 4 ) {
	/* I'm reading last two chars in REMOTE_NAME as user's country code */
	ad_user_geocode0 = *((ad_user_remote_host) + (i-3));
	ad_user_geocode1 = *((ad_user_remote_host) + (i-2));
	ad_user_geocode2 = *((ad_user_remote_host) + (i-1));
      } else {
	ad_user_geocode1 = '2';
	ad_user_geocode2 = '2';
      } /* if strlen remote_host */

      ad_user_domain = ap_psprintf(r->pool, "%c%c", ad_user_geocode1, ad_user_geocode2);

    } /* if reserveddomains */
    
#endif


    /* Get pointer reserved pages table */

    reservedpages_list_base = sconf->ad_reserved_pages;


    /* For each ad_position, select an approriate ad, then generate a block of HTML code for that ad. */

    for (ad_position = 1; ad_position < MAX_AD_POSITIONS; ad_position++) {

      ad_info_ptr1              = NULL;
      ad_info_ptr2              = NULL;
      tmp_ad_text               = NULL;
      ad_reserved_list_ptr1     = NULL;
      ad_number                 = 0;
      ad_numberofads            = sconf->ad_x_numberofads[ad_position];
      rotation_list_base        = sconf->ad_x_rotate[ad_position];	  
      


        /* Is this pagename and ad position on the reserved pages list: */
	
	if ( reservedpages_list_base != NULL ) {
	  
	  tmp_reserved_line = ap_psprintf(r->pool, "%s\t%d\t", tmp_pagename_buffer, ad_position);

#if ABDEBUG
	  tmp_ad_status = ap_psprintf(r->pool, "\n<B>%s: Ad Position %d</B> Looking to see if web page \"%s\" is reserved...\n", 
				    __FILE__, 
				    ad_position,
				    tmp_reserved_line);
	  ap_table_mergen(r->subprocess_env, "ADBANNER_STATUS", tmp_ad_status);
#endif

	  if (( ad_reserved_list_ptr1 = strstr( reservedpages_list_base, tmp_reserved_line ) ) != NULL ) {

	    ad_reserved_list_ptr1 += strlen(tmp_reserved_line);

	    memset(tmp_reserved_adid_offset_buffer, 0, MAX_ADID_OFFSET_SIZE);
	    for (i = 0; ap_isdigit(ad_reserved_list_ptr1[i]); i++) {
	      tmp_reserved_adid_offset_buffer[i] = ad_reserved_list_ptr1[i];
	    }

	    /* Determine the ad_id of the reserved ad we found */

	    ad_offset     = atoi(tmp_reserved_adid_offset_buffer);
	    ad_info_ptr1  = sconf->ad_info_base;
	    ad_info_ptr1 += ad_offset;

	    tmp_ad_status = ap_psprintf(r->pool, "\n<B>%s: Ad Position %d</B> Page and ad_position is reserved: \"%s\", ad_info Offset = %d\n", 
					__FILE__,
					ad_position,
					tmp_reserved_line, 
					ad_offset);
	    ap_table_mergen(r->subprocess_env, "ADBANNER_STATUS", tmp_ad_status);

	  } /* if page is reserved */

	} /* if reserved pages */


#if GEOTARGETING

	/* If we don't have an ad id yet, then check if user's domain and ad position is reserved */
	
	if ( ad_info_ptr1 == NULL ) {

	  if ( reserveddomains_list_base != NULL ) {

	    if ( ap_isalpha( ad_user_geocode2 ) ) {

	      tmp_reserved_line = ap_psprintf(r->pool, "%s\t%d\t", ad_user_domain, ad_position);
	      
#if ABDEBUG
	      tmp_ad_status = ap_psprintf(r->pool, "\n<B>%s: Ad Position %d</B> Looking to see if domain \"%s\" is reserved...\n", 
				    __FILE__, 
				    ad_position,
				    tmp_reserved_line);
	      ap_table_mergen(r->subprocess_env, "ADBANNER_STATUS", tmp_ad_status);
#endif

	      ad_reserved_list_ptr1 = NULL;

	      if (( ad_reserved_list_ptr1 = strstr( reserveddomains_list_base, tmp_reserved_line ) ) != NULL ) {

		ad_reserved_list_ptr1 += strlen(tmp_reserved_line);

		memset(tmp_reserved_adid_offset_buffer, 0, MAX_ADID_OFFSET_SIZE);
		for (i = 0; ap_isdigit(ad_reserved_list_ptr1[i]); i++) {
		  tmp_reserved_adid_offset_buffer[i] = ad_reserved_list_ptr1[i];
		}

		/* Determine the ad_id of the reserved ad we found */

		ad_offset     = atoi(tmp_reserved_adid_offset_buffer);
		ad_info_ptr1  = sconf->ad_info_base;
		ad_info_ptr1 += ad_offset;

		tmp_ad_status = ap_psprintf(r->pool, "\n<B>%s: Ad Position %d</B> User domain and ad_position is reserved: \"%s\", ad_info Offset = %d\n", 
					    __FILE__,
					    ad_position,
					    tmp_reserved_line, 
					    ad_offset);
		ap_table_mergen(r->subprocess_env, "ADBANNER_STATUS", tmp_ad_status);


	      } /* if domain is reserved */
	      
	    } /* if user domain */

	  } /* if reserved domains list */
	  
	} /* if ad_info_ptr1 */

#endif
        /* end if GEOTARGETING */



	/* If we still don't have an ad id, then pick from general rotation list */

	if ( ad_info_ptr1 == NULL ) {

	  if ((rotation_list_base == NULL) && ( ad_reserved_list_ptr1 == NULL )) {
	    tmp_ad_status = ap_psprintf(r->pool, "\n<B>%s: Ad Position %d</B> Position %d rotation list is empty.\n", 
					__FILE__,
					ad_position,
					ad_position);
	    ap_table_mergen(r->subprocess_env, "ADBANNER_STATUS", tmp_ad_status);
	    continue;	/* Goes to next iteration of for{ ad_position 1..5 } loop */
	  }

	  /* Skip ahead to to next ad_position if numberofads < 1 */

	  if (( ad_numberofads <= 0 )  && ( ad_reserved_list_ptr1 == NULL )) {
	    tmp_ad_status = ap_psprintf(r->pool, "\n<B>%s: Ad Position %d</B> Position %d numberofads is 0.\n", 
					__FILE__,
					ad_position,
					ad_position);
	    ap_table_mergen(r->subprocess_env, "ADBANNER_STATUS", tmp_ad_status);
	    continue;	/* Goes to next iteration of for{ ad_position 1..5 } loop */
	  }


#if RANDOM_ADS
	  ad_number = tmp_time % ad_numberofads;
#else
	  ad_number = sconf->ad_x_currentad[ad_position];
#endif


#if ABDEBUG
	  tmp_ad_status = ap_psprintf(r->pool, "\n<B>%s: Ad Position %d</B> Serving general rotation ad, offset %d \n", 
				      __FILE__, 
				      ad_position, 
				      ad_number);
	  ap_table_mergen(r->subprocess_env, "ADBANNER_STATUS", tmp_ad_status);
#endif


	  /* Apply an offset value to the ad_info_base pointer to read the ad_info data string for this ad */

	  ad_offset     = *(rotation_list_base + ad_number);
	  ad_info_ptr1  = sconf->ad_info_base;
	  ad_info_ptr1 += ad_offset;


	}  /* if ad_info_ptr1 */



	/* We know the starting byte of our ad_id. Now read its entire row out of the ad_info table */

	ad_info_ptr2 = strpbrk(ad_info_ptr1, "\n\0");


	/* Subtract pointers to figure out how many bytes to read from this row in ad_info table */

	if (ad_info_ptr2 != NULL) {
	    ad_info_line_length = (ad_info_ptr2 - ad_info_ptr1);
	}
	else {
	    ad_info_line_length = MAX_LINE_LENGTH;
	}

	if (ad_info_line_length >= MAX_LINE_LENGTH) {
	    ad_info_line_length = MAX_LINE_LENGTH;
	}


	/* Copy this ad_info line into a temporary buffer we can trash the string while calling strtok() */

	memset(tmp_ad_info_line_buffer, 0, MAX_LINE_LENGTH);

	for (i = 0; i < ad_info_line_length; i++) {
	    tmp_ad_info_line_buffer[i] = *(ad_info_ptr1 + (i));
	}




#if MEGABANNERS

	/* Override whatever we just did to get the ad_info data if this is a megabanner display portion */
	/* TODO: Here's an evil thought: Serve both megabanner part B and ad 2 : append them together. Hmmmmm.... */

	if ( ( tmp_megabanner_id[4] == '6' ) && ( ad_position == 2 ) && ( ad_reserved_list_ptr1 == NULL )) {

	    memset(tmp_ad_info_line_buffer, 0, MAX_LINE_LENGTH);
	    memcpy(tmp_ad_info_line_buffer, tmp_megabanner_id, ADID_SIZE);

	}

#endif



	/* Make the ADBANNER_STATUS environment variable contain various debugging info */

	tmp_ad_status = ap_psprintf(r->pool, "\n<B>%s: Ad Position %d</B> \n\
            ad module version =  1.05, Apr.12.2001 \n\
          ad_user_remote_host =  %s  \n\
               ad_user_domain =  %s  \n\
                    child pid =  %d  \n\
          sconf->ad_info_base =  %ul \n\
                 ad_info_ptr1 =  %ul \n\
           rotation_list_base =  %ul \n\
          tmp_pagename_buffer = '%s' \n\
              tmp_cachebuster = '%s' \n\
                    ad_number =  %d  \n\
          ad_info_line_length =  %d  \n\
            tmp_megabanner_id = '%s' \n\
      tmp_ad_info_line_buffer = '%s' \n",
				    __FILE__,
				    ad_position,
				    ad_user_remote_host,
				    ad_user_domain,
				    getpid(),
				    sconf->ad_info_base, 
				    ad_info_ptr1,
				    rotation_list_base,
				    tmp_pagename_buffer, 
				    tmp_cachebuster,
				    ad_number,
				    ad_info_line_length, 
				    tmp_megabanner_id, 
				    tmp_ad_info_line_buffer);

	ap_table_mergen(r->subprocess_env, "ADBANNER_STATUS", tmp_ad_status);



	/* The fifth char in the ad_id string tells if this is a rich-media (HTML) ad */

	tmp_ad_type = tmp_ad_info_line_buffer[4];



	/* Grock HTML ad */

	if (tmp_ad_type == '5') {

	    memset(tmp_htmlad_id, 0, ADID_SIZE + 1);
	    memcpy(tmp_htmlad_id, tmp_ad_info_line_buffer, ADID_SIZE);


	    /* Make sure we have a valid path to HTML ad files */


	    if (sconf->ad_htmlad_dir == NULL) {

		tmp_ad_text = ap_psprintf(r->pool, "\n<!-- %s error: Position %d, Ad number %d (%s) : AdBannerHtmlAdDir path is invalid or not specified. --> \n\n", 
					  __FILE__, 
					  ad_position, 
					  ad_number, 
					  tmp_htmlad_id);

	    } else {

		/* Grok text from HTML ad dat file */

		tmp_htmlad_filepath = ap_psprintf(r->pool, "%sad%s.dat", sconf->ad_htmlad_dir, tmp_htmlad_id);

		if ((fd = ap_popenf(r->pool, tmp_htmlad_filepath, O_RDONLY, 0)) < 1) {

		    tmp_ad_text = ap_psprintf(r->pool, "\n<!-- %s: Position %d, Ad number %d (%s) : Can't read %s %s -->\n",
					      __FILE__, 
					      ad_position, 
					      ad_number, 
					      tmp_htmlad_id, 
					      tmp_htmlad_filepath, 
					      strerror(errno));
		} else {


		    i = lseek(fd, 0L, 2);
		    lseek(fd, 0L, 0);
		    tmp_htmlad_filecontents = (char *) ap_pcalloc(r->pool, (i + 1) * (sizeof(char)));
		    read(fd, tmp_htmlad_filecontents, i);
		    ap_pclosef(r->pool, fd);


		    /* Replace INSERT_CACHEBUSTER string with epoch time wherever it appears in HTML ad text */

		    while ((tmp_htmlad_filecontents_ptr1 = strstr(tmp_htmlad_filecontents, "INSERT_CACHEBUSTER")) > 0) {

			/* 
			   TODO: This bit of logic is rather hard-coded as should be made more flexible as far 
			   as length of epoch time string. In Linux tmp_cachebuster = epoch time = 9 characters = 
			   1/2 length of "INSERT_CACHEBUSTER" so I do the string replace for{} loop twice.
			 */

			for (i = 0; i < strlen(tmp_cachebuster); i++, tmp_htmlad_filecontents_ptr1++) {
			    *tmp_htmlad_filecontents_ptr1 = tmp_cachebuster[i];
			}
			for (i = 0; i < strlen(tmp_cachebuster); i++, tmp_htmlad_filecontents_ptr1++) {
			    *tmp_htmlad_filecontents_ptr1 = tmp_cachebuster[i];
			}

		    }  /* while strstr() */


		    /* If the HTML ad layer file does not have an impression counter then add it */

		    tmp_htmlad_imageurl = ap_psprintf(r->pool, "%s%s", sconf->ad_images_url, tmp_htmlad_id);

		    if (strstr(tmp_htmlad_filecontents, tmp_htmlad_imageurl) == NULL) {

		      tmp_ad_text = ap_psprintf(r->pool,
						"%s\n<TABLE BORDER=0 CELLPADDING=0 CELLSPACING=0><TR><TD WIDTH=\"1\"><SCRIPT LANGUAGE=\"JAVASCRIPT\">\n\
<!--\n\
now = new Date();\n\
tail = now.getTime();\n\
document.write(\"<IMG SRC=\'%s.gif\?\");\n\
document.write(tail);\n\
document.write(\"\' WIDTH=1 HEIGHT=1 BORDER=0><BR>\");\n\
//-->\n\
</SCRIPT>\n\
<NOSCRIPT>\n\
<IMG SRC=\"%s.gif\?%s\" WIDTH=1 HEIGHT=1 BORDER=0><BR>\n\
</NOSCRIPT></TD></TR></TABLE>\n",
						tmp_htmlad_filecontents,
						tmp_htmlad_imageurl,
						tmp_htmlad_imageurl,
						tmp_cachebuster);


		    } else {

		      tmp_ad_text = tmp_htmlad_filecontents;

		    }  /* if no img tag  */
 

		}   /* if openf tmp_htmlad_filepath */


	    }	  /* if ad_htmlad_dir exists */



#if MEGABANNERS


	} else if (tmp_ad_type == '6') {

	    /* If megabanner ID was set already then don't reset it */

	    if (tmp_megabanner_id[0] == 0) {
		memcpy(tmp_megabanner_id, tmp_ad_info_line_buffer, ADID_SIZE);
	    }

	    /* Make sure we have a valid path to HTML ad files */

	    if (sconf->ad_htmlad_dir == NULL) {

		tmp_ad_text = ap_psprintf(r->pool, "\n<!-- %s error: Position %d, Ad number %d (%s) : AdBannerHtmlAdDir path is invalid or not specified. -->\n\n", 
					  __FILE__, 
					  ad_position, 
					  ad_number, 
					  tmp_megabanner_id);

	    } else {

		/* If large banner is a 6000 ("megabanner") then position 2 ad must be it's partner component */

		if (ad_position == 1) {

		    tmp_htmlad_filepath = ap_psprintf(r->pool, "%sad%s.dat", sconf->ad_htmlad_dir, tmp_megabanner_id);

		} else if (ad_position == 2) {

		    tmp_htmlad_filepath = ap_psprintf(r->pool, "%sad%s-b.dat", sconf->ad_htmlad_dir, tmp_megabanner_id);

		}

		if ((fd = ap_popenf(r->pool, tmp_htmlad_filepath, O_RDONLY, 0)) < 1) {

		    tmp_ad_text = ap_psprintf(r->pool, "\n<!-- %s: Position %d, Ad number %d (%s, megabanner) : Can't read %s %s --> \n",
					      __FILE__, 
					      ad_position, 
					      ad_number, 
					      tmp_megabanner_id, 
					      tmp_htmlad_filepath, 
					      strerror(errno));
		} else {


		    i = lseek(fd, 0L, 2);
		    lseek(fd, 0L, 0);
		    tmp_htmlad_filecontents = (char *) ap_pcalloc(r->pool, (i + 1) * (sizeof(char)));
		    read(fd, tmp_htmlad_filecontents, i);
		    ap_pclosef(r->pool, fd);


		    /* Replace INSERT_CACHEBUSTER string with epoch time wherever it appears in HTML ad text */

		    while ((tmp_htmlad_filecontents_ptr1 = strstr(tmp_htmlad_filecontents, "INSERT_CACHEBUSTER")) > 0) {

			/* 
			   TODO: This bit of logic is rather hard-coded as should be made more flexible as far 
			   as length of epoch time string. In Linux tmp_cachebuster = epoch time = 9 characters = 
			   1/2 length of "INSERT_CACHEBUSTER" so I do the string replace for{} loop twice.
			 */

			for (i = 0; i < strlen(tmp_cachebuster); i++, tmp_htmlad_filecontents_ptr1++) {
			    *tmp_htmlad_filecontents_ptr1 = tmp_cachebuster[i];
			}
			for (i = 0; i < strlen(tmp_cachebuster); i++, tmp_htmlad_filecontents_ptr1++) {
			    *tmp_htmlad_filecontents_ptr1 = tmp_cachebuster[i];
			}

		    }		/* if strstr() */


		    if (ad_position == 1) {

		      /* If the position 1 portion of the megabanner file does not have an impression counter then add it */

		      tmp_htmlad_imageurl = ap_psprintf(r->pool, "%s%s", sconf->ad_images_url, tmp_megabanner_id);

		      if (strstr(tmp_htmlad_filecontents, tmp_htmlad_imageurl) == NULL) {

			tmp_ad_text = ap_psprintf(r->pool,
						"%s\n<TABLE BORDER=0 CELLPADDING=0 CELLSPACING=0><TR><TD WIDTH=\"1\"><SCRIPT LANGUAGE=\"JAVASCRIPT\">\n\
<!--\n\
now = new Date();\n\
tail = now.getTime();\n\
document.write(\"<IMG SRC=\'%s.gif\?\");\n\
document.write(tail);\n\
document.write(\"\' WIDTH=1 HEIGHT=1 BORDER=0><BR>\");\n\
//-->\n\
</SCRIPT>\n\
<NOSCRIPT>\n\
<IMG SRC=\"%s.gif\?%s\" WIDTH=1 HEIGHT=1 BORDER=0><BR>\n\
</NOSCRIPT></TD></TR></TABLE>\n",
						  tmp_htmlad_filecontents,
						  tmp_htmlad_imageurl,
						  tmp_htmlad_imageurl,
						  tmp_cachebuster);

		      } else {

			tmp_ad_text = tmp_htmlad_filecontents;

		      }  /* if no img tag  */


		    } else if (ad_position == 2) {

		      tmp_ad_text = tmp_htmlad_filecontents;

		    }


		}  /* if openf tmp_htmlad_filepath */

	    }	/* if ad_htmlad_dir */


#endif


        /* Normal plain GIF ad banner */

	} else {

	    /* Reset each char pointer to NULL just to be safe */

	    tmp_ad_banner_adid = NULL;
	    tmp_ad_banner_height = NULL;
	    tmp_ad_banner_width = NULL;
	    tmp_ad_banner_target_url = NULL;
	    tmp_ad_banner_alt_text = NULL;

	    /* Read each tab-separated token from ad_info_line */

	    tmp_ad_banner_adid = strtok(tmp_ad_info_line_buffer, "\t");
	    tmp_ad_banner_height = strtok(NULL, "\t");
	    tmp_ad_banner_width = strtok(NULL, "\t");
	    tmp_ad_banner_target_url = strtok(NULL, "\t");
	    tmp_ad_banner_alt_text = strtok(NULL, "\t");


	    /* Grock together the HTML code for a normal GIF ad banner */

	    tmp_ad_text = ap_psprintf(r->pool,
				      "<TABLE BORDER=0 CELLPADDING=0 CELLSPACING=0><TR><TD WIDTH=%s><SCRIPT LANGUAGE=\"JAVASCRIPT\">\n\
<!--\n\
now = new Date();\n\
tail = now.getTime();\n\
AltText = \"\\\"%s\\\"\";\n\
document.write(\"<A HREF=\'%s\?%s,%s\' TARGET=_top>\");\n\
document.write(\"<IMG SRC=\'%s%s.gif\?\");\n\
document.write(tail);\n\
document.write(\"\' WIDTH=\%s HEIGHT=%s BORDER=0 ALT=\");\n\
document.write(AltText);\n\
document.write(\"></A><BR>\");\n\
//-->\n\
</SCRIPT>\n\
<NOSCRIPT>\n\
<A HREF=\"%s\?%s,%s\" TARGET=_top><IMG SRC=\"%s%s.gif\?%s\" WIDTH=%s HEIGHT=%s BORDER=0 ALT=\"%s\"></A><BR>\n\
</NOSCRIPT></TD></TR></TABLE>\n",
				      tmp_ad_banner_width, 
				      tmp_ad_banner_alt_text, 
				      sconf->ad_click_url, 
				      tmp_pagename_buffer, 
				      tmp_ad_banner_adid, 
				      sconf->ad_images_url, 
				      tmp_ad_banner_adid, 
				      tmp_ad_banner_width, 
				      tmp_ad_banner_height, 
				      sconf->ad_click_url, 
				      tmp_pagename_buffer, 
				      tmp_ad_banner_adid,
				      sconf->ad_images_url, 
				      tmp_ad_banner_adid, 
				      tmp_cachebuster,
				      tmp_ad_banner_width, 
				      tmp_ad_banner_height, 
				      tmp_ad_banner_alt_text);


	}			/* if ad_type == 5 */


#if RANDOM_ADS
#else

	/* Increment the round-robin ad rotation list counter */

	ad_number++;
	if (ad_number >= ad_numberofads) {
	    ad_number = 0;
	}

#endif



	/* Store pointers to the next ad to serve */


#if RANDOM_ADS
#else
	sconf->ad_x_currentad[ad_position] = ad_number;
#endif



	switch (ad_position) {

	case 1:

	  tmp_ad_env_label = "AD_BANNER_1";

	  /* Include the pagecount tags above position 1 banner */

	  if ( tmp_extra_text ) {
#if PAGECOUNTGIF_BELOW_AD
	    tmp_ad_text = ap_psprintf(r->pool, "%s%s", tmp_ad_text, tmp_extra_text);
#else
	    tmp_ad_text = ap_psprintf(r->pool, "%s%s", tmp_extra_text, tmp_ad_text);
#endif
	  }
	  break;



	case 2:

	    tmp_ad_env_label = "AD_BANNER_2";


#if MEGABANNERS

	    if ( sconf->ad_navbar_html != NULL ) {

	      if (tmp_ad_type == '6') {

		/* Megabanner w/ navbar: Put network navbar below Magabanner Part B */

		tmp_ad_text = ap_psprintf(r->pool, 
"\n<IMG SRC=\"%s/top_mega.gif\" WIDTH=135 HEIGHT=10><BR>\n\
<TABLE CELLPADDING=0 CELLSPACING=0 WIDTH=135 BORDER=0>\
<TR><TD BGCOLOR=\"#%s\" WIDTH=135 ALIGN=CENTER VALIGN=TOP>\
<FONT FACE=\"ARIAL,HELVETICA\" SIZE=2> %s </FONT>\n\
</TD></TR></TABLE><IMG SRC=\"%s/bottom.gif\" WIDTH=135 HEIGHT=10><BR><BR>\n %s \n", 
					sconf->ad_navbar_gifsurl, 
					sconf->ad_navbar_bgcolor, 
					tmp_ad_text, 
					sconf->ad_navbar_gifsurl, 
					sconf->ad_navbar_html);
	      
	      } else {

		/* Non-megabanner w/ navbar: Put network navbar above the second ad */

		tmp_ad_text = ap_psprintf(r->pool, 
"\n %s \n<IMG SRC=\"%s/top.gif\" WIDTH=135 HEIGHT=10><BR>\n\
<TABLE CELLPADDING=0 CELLSPACING=0 WIDTH=135 BORDER=0>\
<TR><TD BGCOLOR=\"#%s\" WIDTH=135 ALIGN=CENTER VALIGN=TOP>\
<FONT FACE=\"ARIAL,HELVETICA\" SIZE=2> %s </FONT>\n\
</TD></TR></TABLE><IMG SRC=\"%s/bottom.gif\" WIDTH=135 HEIGHT=10><BR><BR>\n\n", 
					sconf->ad_navbar_html, 
					sconf->ad_navbar_gifsurl, 
					sconf->ad_navbar_bgcolor, 
					tmp_ad_text, 
					sconf->ad_navbar_gifsurl);


	      }  /* if ad_type == 6  */

	    } /* if sconf->ad_navbar_html */

#endif

	    break;






	case 3:

	    tmp_ad_env_label = "AD_BANNER_3";
	    break;


	case 4:

	    tmp_ad_env_label = "AD_BANNER_4";
	    break;


	case 5:

	    tmp_ad_env_label = "AD_BANNER_5";
	    break;


	}  /* switch ad_position */



	/* Grok our ad banner HTML into an ENV variable */

	if (tmp_ad_text != NULL) {
	    ap_table_setn(r->subprocess_env, tmp_ad_env_label, tmp_ad_text);
	}



    }  /* for ad_position = 1..5 */



    /* 
       We're done processing this http request. "return DECLINED" means we're 
       passing control of this request back to Apache 
     */

    return DECLINED;


}  /* sub adbanners_rotate */








/*
   This bit of code tells Apache which of our own functions to call when the server starts up
   and reads our directives from the httpd.conf file. If the directive type is one of ours
   then Apache reads the argument immediately following the directive and passes that 
   string as a char* argument to our directive handler function.
 */


static const command_rec adbanners_module_cmds[] =
{
    {"AdBannerPagecountURL", adbanners_module_read_config_pagecounturl, NULL,
     RSRC_CONF, TAKE1, "a relative or full URL to directory where the primary pagecount gif for this VirtualHost."},
    {"AdBannerPagecountBackupURL", adbanners_module_read_config_pagecountbackupurl, NULL,
     RSRC_CONF, TAKE1, "a relative or full URL to directory where the secondary pagecount gif for this VirtualHost."},
    {"AdBannerClickURL", adbanners_module_read_config_clickurl, NULL,
     RSRC_CONF, TAKE1, "a relative or full URL to the ad click redirector script for this VirtualHost."},
    {"AdBannerImagesBaseURL", adbanners_module_read_config_imagesurl, NULL,
     RSRC_CONF, TAKE1, "a relative or full URL to directory where ad banner images are stored."},
    {"AdBannerHtmlAdDir", adbanners_module_read_config_htmladdir, NULL,
     RSRC_CONF, TAKE1, "full path to a directory from which HTML source files for each HTML ad can be read from."},
    {"AdBannerInfoFile", adbanners_module_read_config_adinfofile, NULL,
     RSRC_CONF, TAKE1, "full path of the file where ad_info table for ad banners is stored."},
    {"AdBannerRotateFile", adbanners_module_read_config_rotatelist, NULL,
     RSRC_CONF, TAKE2, "ad position number followed by the full path to the rotation list file for the ad banners in that position."},
    {"AdBannerReservedPagesFile", adbanners_module_read_config_reservedlist, NULL,
     RSRC_CONF, TAKE1, "full path to a file that lists reserved web pages and the ad ID for each reserved page."},

#if GEOTARGETING
    {"AdBannerReservedDomainsFile", adbanners_module_read_config_reserveddomains_list, NULL,
     RSRC_CONF, TAKE1, "full path to a file that lists reserved domains and the ad ID for each reserved domain."},
#endif

#if MEGABANNERS
    {"AdBannerNavbar", adbanners_module_read_config_navbarinfo, NULL,
     RSRC_CONF, TAKE3, "page background color hex code, followed by the URL of the decorative GIFs server,\
followed by the full path to a file that contains the HTML to include with position 2 ads."},
#endif

    {NULL},
};




/*
   Normally this bit of code appears at the end of an Apache module. It tells Apache which of our own
   functions to call during each given stage of Apache's task list, whether it be when starting up
   (reading the httpd.conf file) or when handling requests, or whenever a certain event is occuring.
 */


module MODULE_VAR_EXPORT adbanners_module =
{
    STANDARD_MODULE_STUFF,
    NULL,			        /* initializer */
    NULL,			        /* dir config creater */
    NULL,			        /* dir merger --- default is to override */
    create_adbanners_server_config,	/* server config */
    merge_adbanners_server_config,	/* merge server configs */
    adbanners_module_cmds,	        /* command table */
    NULL,			        /* handlers (adbanners_handler[]) */
    NULL,			        /* filename translation */
    NULL,			        /* check_user_id */
    NULL,			        /* check auth */
    NULL,			        /* check access */
    NULL,			        /* type_checker */
    NULL,			        /* fixups */
    NULL,			        /* logger */
    NULL,			        /* header parser */
    adbanners_child_init,	        /* child_init */
    NULL,			        /* child_exit */
    adbanners_rotate,		        /* post read-request */
};
